"""Implementation and optimization of React agent."""

from typing import List, Union, Callable, Optional, Any, Dict
from dataclasses import dataclass, field
from adalflow.core.base_data_class import DataClass
import logging
import traceback


from adalflow.core.generator import Generator
from adalflow.optim.grad_component import GradComponent
from adalflow.optim.parameter import Parameter, ParameterType
from adalflow.core.func_tool import FunctionTool, AsyncCallable
from adalflow.core.tool_manager import ToolManager
from adalflow.core.component import Component
from adalflow.components.output_parsers import JsonOutputParser
from adalflow.core.types import (
    StepOutput,
    GeneratorOutput,
    Function,
    FunctionOutput,
)
from adalflow.optim.grad_component import fun_to_grad_component
from adalflow.core.model_client import ModelClient
from adalflow.utils.logger import printc


log = logging.getLogger(__name__)

__all__ = ["DEFAULT_REACT_AGENT_SYSTEM_PROMPT", "ReActAgent"]


react_agent_task_desc = r"""
<START_OF_TASK_SPEC>
You are an excellent task planner.
Answer the input query using the tools provided below with maximum accuracy.

Each step you will read the previous thought, Action(name, kwargs), and Observation(execution result of the action) and then provide the next Thought and Action.

Follow function docstring to best call the tool.
- For simple queries: Directly call the ``finish`` action and provide the answer.
- For complex queries:
    - Step 1: Read the user query and divide it into multisteps. Start with the first tool/subquery.
    - Call one tool at a time to solve each subquery/subquestion. \
    - At step 'finish', give the final answer based on all previous steps.
REMEMBER:
- Action MUST call one of the tools. It CANNOT be empty.
- You will ALWAYS END WITH 'finish' tool to finish the task directly with answer or failure message.
- When the tool is a class method and when class_instance exists, use <class_instance_value>.<func_name> to call instead (NOT the CLASS NAME)
<END_OF_TASK_SPEC>
"""

# - In this case, you are working as a multi-hop retriever and your answer in finish MUST be verbatim short factoid responses from retrieved context.
# - Answer with only the exact answer phrase, not a full sentence or paragraph.

DEFAULT_REACT_AGENT_SYSTEM_PROMPT = r"""<START_OF_SYSTEM_PROMPT>
{{react_agent_task_desc}}
- You cant use more than {{max_steps}} steps. At the {{max_steps}}th current step, must finish with answer.

{# Tools #}
{% if tools %}
<START_OF_TOOLS>
Tools and instructions:
{% for tool in tools %}
{{ loop.index }}.
{{tool}}
------------------------
{% endfor %}
<END_OF_TOOLS>
{% endif %}
{# Context Variables #}
{% if context_variables is not none %}
<START_OF_CONTEXT>
You have access to context_variables with the following keys:
{% for key, value in context_variables.items() %}
{{ key }}
------------------------
{% endfor %}
You can either pass context_variables or context_variables['key'] to the tools depending on the tool's requirements.
<END_OF_CONTEXT>
{% endif %}
{# output format and examples for output format #}
<START_OF_OUTPUT_FORMAT>
{{output_format_str}}
<END_OF_OUTPUT_FORMAT>
{% if examples %}
<START_OF_EXAMPLES>
Examples:
{% for example in examples %}
{{example}}
------------------------
{% endfor %}
<END_OF_EXAMPLES>
{% endif %}
<END_OF_SYSTEM_PROMPT>
-----------------
<START_OF_USER_QUERY>
Input query:
{{ input_str }}
_____________________
Current Step/Max Step: {{step_history|length + 1}} / {{max_steps}}
{# Step History #}
{% if step_history %}
<STEPS>
Your previous steps:
{% for history in step_history %}
Step {{ loop.index }}.
{% if history.action %}
"thought": "{{history.action.thought}}",
"name": "{{history.action.name}},
"kwargs": {{history.action.kwargs}}",
{% endif %}
"Observation": "{{history.observation}}"

------------------------
{% endfor %}
</STEPS>
{% endif %}
<END_OF_USER_QUERY>
"""


class CombineStepHistory(GradComponent):
    def __init__(self):
        super().__init__(desc="Extract the final answer from the step history.")

    def call(
        self,
        step_history: List[StepOutput],
        react_agent_task_desc: str,  # skip connection
        id: Optional[str] = None,
    ) -> str:
        if not step_history:
            return ""
        answer = step_history[-1].observation
        return answer


@dataclass
class ReActOutput(DataClass):
    r"""Similar to GeneratorOutput, but with additional step history and final answer."""

    id: Optional[str] = field(
        default=None, metadata={"desc": "The unique id of the output"}
    )
    step_history: List[StepOutput] = field(
        metadata={"desc": "The history of steps."}, default_factory=list
    )

    answer: Any = field(metadata={"desc": "The final answer."}, default=None)


class ReActAgent(Component):
    __doc__ = r"""ReActAgent uses generator as a planner that runs multiple and sequential functional call steps to generate the final response.
    The planner will generate a Function data class as action for each step that includes a "thought" field.
    The execution result is stored in the "observation" field of the StepOutput data class.
    If the execution failed, it will store the error message in the "observation" field so that we can auto-optimize it to correct the error.

    The final answer can be different in training and eval mode:
    - Training: the final answer will be
    Users need to set up:
    - tools: a list of tools to use to complete the task. Each tool is a function or a function tool.
    - max_steps: the maximum number of steps the agent can take to complete the task.
    - use_llm_as_fallback: a boolean to decide whether to use an additional LLM model as a fallback tool to answer the query.
    - model_client: the model client to use to generate the response.
    - model_kwargs: the model kwargs to use to generate the response.
    - template: the template to use to generate the prompt. Default is DEFAULT_REACT_AGENT_SYSTEM_PROMPT.
    - context_variables: the context variables to use in the prompt.
    - use_cache: a boolean to decide whether to use the cache to store the generated responses for the planner.
    - debug: a boolean to decide whether to print debug information.

    For the generator, the default arguments are:
    (1) default prompt: DEFAULT_REACT_AGENT_SYSTEM_PROMPT
    (2) default output_processors: JsonParser

    There are `examples` which is optional, a list of string examples in the prompt.

    Example:

    .. code-block:: python

        from core.openai_client import OpenAIClient
        from components.agent.react import ReActAgent
        from core.func_tool import FunctionTool
        # define the tools
        def multiply(a: int, b: int) -> int:
            '''Multiply two numbers.'''
            return a * b
        def add(a: int, b: int) -> int:
            '''Add two numbers.'''
            return a + b
        agent = ReActAgent(
            tools=[multiply, add],
            model_client=OpenAIClient(),
            model_kwargs={"model": "gpt-3.5-turbo"},
        )

        # Using examples:

        call_multiply = FunctionExpression.from_function(
            thought="I want to multiply 3 and 4.",



    Reference:
    [1] https://arxiv.org/abs/2210.03629, published in Mar, 2023.
    """

    # TODO: allow users to pass in a few examples. Need to be a list of FunctionExpression instances.
    def __init__(
        self,
        # added arguments specifc to React
        tools: List[Union[Callable, AsyncCallable, FunctionTool]] = [],
        max_steps: int = 10,
        add_llm_as_fallback: bool = True,
        # TODO: the examples are just for specifying the output format, not end to end input-output examples, need further optimization
        examples: Union[List[Function], List[str]] = [],
        *,
        # the following arguments are mainly for the planner
        model_client: ModelClient,
        model_kwargs: Dict = {},
        # template for the planner
        template: Optional[str] = None,  # allow users to customize the template
        context_variables: Optional[Dict] = None,  # context variables
        use_cache: bool = True,
        debug: bool = False,
    ):
        super().__init__()
        template = template or DEFAULT_REACT_AGENT_SYSTEM_PROMPT

        self.max_steps = max_steps

        self.add_llm_as_fallback = add_llm_as_fallback
        self.context_variables = context_variables
        self.debug = debug
        self.use_cache = use_cache

        processed_tools = self._init_tools(tools, model_client, model_kwargs)
        self.tool_manager: ToolManager = ToolManager(
            tools=processed_tools,
            additional_context={"context_variables": self.context_variables},
        )

        ouput_data_class = Function
        example = Function(
            thought="Based on all the subtasks, I am able to answer the question. Following the finish doc string, and ....",
            name="finish",
            kwargs={"answer": "final answer"},
        )
        self._examples = examples + [example]

        output_parser = JsonOutputParser(
            data_class=ouput_data_class,
            examples=self._examples,
            return_data_class=True,
            include_fields=[
                "thought",
                "name",
                "kwargs",
            ],
        )
        prompt_kwargs = {
            "tools": self.tool_manager.yaml_definitions,
            "output_format_str": output_parser.format_instructions(),
            "react_agent_task_desc": Parameter(
                name="react_agent_task_desc",
                data=react_agent_task_desc,
                # data="You are an excellent task planner. Answer the input query using the tools provided below with maximum accuracy.\n\nEach step you will read the previous thought, Action(name, kwargs), and Observation(execution result of the action) and then provide the next Thought and Action.\n\n<START_OF_TASK_SPEC>\nFollow function docstring to best call the tool.\n- For simple queries: Directly call the 'finish' action and answer with a concise 'yes' or 'no' when it fits.\n- For complex queries:\n    - Step 1: Understand the main subject(s) and context of the user query accurately.\n    - Step 2: Break down the query into multisteps, starting with the first tool/subquery.\n    - Ensure each step accurately reflects the subjects under consideration.\n    - Continuously verify your extracted information and logic for factual accuracy using concise comparisons.\n    - At step 'finish', conclude with a precise final answer.\nREMEMBER:\n- Action MUST call one of the tools. It CANNOT be empty.\n- You will ALWAYS END WITH 'finish' tool to conclude the task directly with an answer or failure message.\n- When the tool is a class method and when class_instance exists, use <class_instance_value>.<func_name> to call instead (NOT the CLASS NAME).\n<END_OF_TASK_SPEC>",
                role_desc="Task instruction for the agent to plan steps to solve a question in sequential and multi-steps to get the final answer. \
                For optimizer: you need to adapt this to the current specific task.",
                param_type=ParameterType.PROMPT,
                requires_opt=True,
            ),
            # "examples": Parameter(
            #     name="examples",
            #     data=None,
            #     role_desc="Examples for the ReAct agent.",
            #     param_type=ParameterType.DEMOS,
            #     requires_opt=True,
            # ),
            "context_variables": self.context_variables,
            "max_steps": self.max_steps,
        }
        self.planner = Generator(
            template=template,
            prompt_kwargs=prompt_kwargs,
            output_processors=output_parser,
            model_client=model_client,
            model_kwargs=model_kwargs,
            use_cache=use_cache,
        )

        # besides of form the final output, it adds a skip connection to the planner task description prompt
        self.combine_step_history = CombineStepHistory()

    def _init_tools(
        self,
        tools: List[Union[Callable, AsyncCallable, FunctionTool]],
        model_client: ModelClient,
        model_kwargs: Dict,
    ):
        r"""Initialize the tools. Using reference or else with (copy or deepcopy) we can not set the training/eval mode for each tool."""
        processed_tools = []
        _additional_llm_tool = (
            Generator(model_client=model_client, model_kwargs=model_kwargs)
            if self.add_llm_as_fallback
            else None
        )

        def llm_tool(input: str, **kwargs) -> str:
            """I answer any input query with llm's world knowledge. Use me as a fallback tool or when the query is simple."""
            try:
                output: GeneratorOutput = _additional_llm_tool(
                    prompt_kwargs={"input_str": input}
                )
                response = output.data if output else None
                return response
            except Exception as e:
                log.error(f"Error using the llm_tool: {e}")
                print(f"Error using the llm_tool: {e}")

            return None

        # always add **kwargs for us to track the id, __doc__ as the predecessors.
        from adalflow.optim.grad_component import fun_to_grad_component

        @fun_to_grad_component(
            desc="Finish",
            doc_string=Parameter(
                data="Finish the task with the final answer in the kwargs.",
                param_type=ParameterType.PROMPT,
                requires_opt=True,
                role_desc="Instruct the agent on how to create the final answer from the step history.",
                name="doc_string",
            ),
        )
        def finish(answer: str, **kwargs) -> str:
            return answer

        self._finish = FunctionTool(fn=finish, component=finish)
        processed_tools = tools.copy()
        if self.add_llm_as_fallback:
            processed_tools.append(llm_tool)
        processed_tools.append(self._finish)
        return processed_tools

    def _execute_action(
        self,
        step_output: StepOutput,
        response: Union[Parameter, GeneratorOutput],
        id: Optional[str] = None,
    ) -> Optional[StepOutput]:
        """Parse the action string to a function call and execute it. Update the step_output with the result."""

        def handle_error(response: Parameter, e: str):

            @fun_to_grad_component()
            def set_step_output_with_error(
                step_output: StepOutput, error: str, response: Any
            ):
                """Set the step_output with error."""
                step_output.observation = f"error: {error} at {response.data}"
                return step_output

            response.add_successor_map_fn(
                successor=set_step_output_with_error, map_fn=lambda x: x.data
            )
            return set_step_output_with_error.forward(step_output, e, response)

        step = step_output.step

        if isinstance(response, Parameter):

            try:

                step_output.action = response.data.data
                if self.debug:
                    printc(
                        f"Step test train:  {step}: {step_output.action}", color="blue"
                    )

                if isinstance(response.data.data, Function):
                    response.data.data.kwargs.update({"id": id})

                result: Union[Parameter, str] = self.tool_manager(
                    expr_or_fun=response,
                    step="execute",
                    map_fn=lambda x: x.data.data,  # Function
                )

                if isinstance(result, str):

                    @fun_to_grad_component()
                    def set_step_output_with_error(step_output: StepOutput, data: str):
                        """Set the step_output with error."""
                        step_output.observation = f"Error {data} in executing action."

                        return step_output

                    response.add_successor_map_fn(
                        successor=set_step_output_with_error,
                        map_fn=lambda x: x.data.data,
                    )
                    step_output = set_step_output_with_error.forward(
                        step_output, response
                    )

                    return step_output

            except Exception as e:
                e = f"{e} Error executing action: {response.data}"
                return handle_error(response, e)

            try:

                step_output.step = step
                step_output.observation = result.data.output

                # update the execution result to the step_output to be consistent with the eval version
                result.data = step_output
                result.role_desc = "The result of the action execution, observation is the final answer"
                result.param_type = ParameterType.OUTPUT
                return result
            except Exception as e:
                e = f"{e} Error converting function output to step output: {result.data}"

                return handle_error(response, e)

        else:

            return self._execute_action_eval_mode(
                x=response,
                step_output=step_output,
                step=step,
                id=id,
            )

    def _execute_action_eval_mode(
        self,
        x: GeneratorOutput,
        step_output: StepOutput,
        step: int,
        id=None,
    ) -> StepOutput:
        """Execute the action and update the step_output."""
        if x.error or not x.data:
            error_msg = f"Error planning step {step}: {x.error}"
            step_output.observation = error_msg
            step_output.action = None
            log.error(error_msg)
            return step_output
        else:
            try:
                fun_expr: Function = x.data

                step_output.action = fun_expr
                # # add id to the function
                fun_expr.kwargs.update({"id": id})

                if step_output and step_output.action:

                    result: FunctionOutput = self.tool_manager(
                        expr_or_fun=x.data,  # Function
                        step="execute",
                    )

                    step_output.observation = result.output
                    if self.debug:
                        printc(f"Step {step}: \n{step_output}\n_______\n", color="blue")
                    return step_output
                else:
                    if self.debug:
                        printc(f"Failed to parse response for step {step}", color="red")
                    log.error(f"Failed to parse response for step {step}")
                    return step_output
            except Exception as e:
                error_msg = f"Error parsing response for step {step}: {e}"
                step_output.observation = error_msg
                log.error(error_msg)
                if self.debug:
                    printc(error_msg, color="red")
                return step_output

    def _run_one_step(
        self,
        step: int,
        prompt_kwargs: Dict,
        model_kwargs: Dict,
        id: Optional[str] = None,
        step_history: List[StepOutput] = [],
    ) -> Union[Parameter, StepOutput]:
        """Run one step of the agent. Plan and execute the action for the step.
        Need to deal with both train and eval mode on the self.planner.
        """
        if self.debug:
            printc(f"step: {step}", color="yellow")

        step_history_value = []
        for step_output in step_history:
            if isinstance(step_output, Parameter):
                step_history_value.append(step_output.data)
            else:
                step_history_value.append(step_output)

        prompt_kwargs["step_history"] = step_history_value

        for data in step_history_value:
            if not data:
                raise ValueError(
                    f"Expected StepOutput, but got {type(data)}, all steps: {step_history_value}"
                )
            if not isinstance(data, StepOutput):
                raise ValueError(
                    f"Expected StepOutput, but got {type(data)}, all steps: {step_history_value}"
                )

        log.debug(
            f"Running step {step} with prompt: {self.planner.prompt(**prompt_kwargs)}"
        )
        try:

            response: Union[GeneratorOutput, Parameter] = self.planner(
                prompt_kwargs=prompt_kwargs, model_kwargs=model_kwargs, id=id
            )
            # prompt_str = self.planner.get_prompt(**prompt_kwargs)
            # printc(f"Prompt: {prompt_str}", color="yellow")

        except Exception as e:
            error_msg = f"Error happened in planner response at step {step}: {e}.\n"
            error_msg += (
                f"Prompt kwargs: {prompt_kwargs}\nModel kwargs: {model_kwargs}\n"
            )
            error_msg += f"Traceback:\n{traceback.format_exc()}"
            raise RuntimeError(error_msg)

        step_output: StepOutput = StepOutput(step=step)

        try:

            if self.training and isinstance(response, Parameter):

                if not isinstance(response.data, GeneratorOutput):
                    raise ValueError(
                        f"Expected GeneratorOutput, but got {type(response.data)}, value: {response.data}"
                    )
                # Detect planner parsing errors to FunctionExpression so that the prompt can be trained to self-correct
                if not isinstance(response.data.data, Function):

                    @fun_to_grad_component()
                    def set_step_output_with_error(
                        step_output: StepOutput, data: GeneratorOutput
                    ):
                        """Set the step_output with error."""
                        step_output.observation = f"Error {data.error} in parsing response: {data.raw_response}, data type: {type(data.data)}"
                        return step_output

                    response.add_successor_map_fn(
                        successor=set_step_output_with_error,
                        map_fn=lambda x: x.data,
                    )
                    step_output = set_step_output_with_error.forward(
                        step_output, response
                    )

                else:

                    step_output: Parameter = self._execute_action(
                        step_output, response, id
                    )
                    if not isinstance(step_output, Parameter):
                        raise ValueError(
                            f"Expected Parameter, but got {type(step_output)}, value: {step_output}"
                        )
                if self.debug:
                    printc(f"step_output: {step_output.data}", color="red")
                if not isinstance(step_output, Parameter):
                    raise ValueError(
                        f"Ensure step_output to be Parameter at training mode. Got {type(step_output)}.\n\
                            Please check the observation for error details: {step_output}"
                    )

                return step_output

            else:

                step_output: StepOutput = self._execute_action(
                    step_output=step_output, response=response, id=id
                )
                if not step_output:
                    raise RuntimeError(
                        f"Error executing action at step {step}: {step_output}"
                    )

                if self.debug:
                    printc(f"step_output: {step_output}", color="red")
                return step_output
        except Exception as e:
            error_msg = f"Error during execution at step {step}: {e}.\n"
            error_msg += f"Step output: {step_output}\nResponse: {response}\n"
            error_msg += f"Traceback:\n{traceback.format_exc()}"
            raise RuntimeError(error_msg)

    def _check_last_step(
        self, step_history: List[Union[StepOutput, Parameter]]
    ) -> bool:
        """Check if the last step is the finish step."""
        if not step_history:
            return True

        last_step: Union[StepOutput, Parameter] = step_history[-1]

        if isinstance(last_step, Parameter):
            last_step = last_step.data

        if (
            last_step
            and last_step.action
            and hasattr(last_step.action, "name")
            and last_step.action.name == "finish"
        ):
            return True
        return False

    def _get_answer(
        self, step_history: List[Union[StepOutput, Parameter]]
    ) -> Union[str, "Parameter"]:
        """Get the final answer from the step history.

        When in training mode, we pass the whole step_history to the backward engine to find the feedback
        """
        if not step_history:
            return None

        last_step: Union[StepOutput, Parameter] = step_history[-1]
        if isinstance(last_step, Parameter):

            answer = self.combine_step_history(
                step_history=step_history,
                id=last_step.data_id,
                react_agent_task_desc=self.planner.prompt_kwargs[
                    "react_agent_task_desc"
                ],
            )
            return answer
        else:
            return last_step.observation

    def call(self, *args, **kwargs) -> ReActOutput:
        output = self.bicall(*args, **kwargs)
        if not isinstance(output, ReActOutput) or not output:
            raise ValueError(f"Expected ReActOutput, but got {type(output)}")
        return output

    def forward(self, *args, **kwargs) -> Parameter:
        return self.bicall(*args, **kwargs)

    def _is_step_output_last_step(self, step_output: StepOutput) -> bool:
        """Check if the step output is the last step."""
        step_output_data = (
            step_output.data if isinstance(step_output, Parameter) else step_output
        )
        if (
            step_output_data
            and step_output_data.function
            and step_output_data.function.name == "finish"
        ):
            return True
        return False

    def bicall(
        self,
        input: str,
        promt_kwargs: Optional[Dict] = {},
        model_kwargs: Optional[Dict] = {},
        id: Optional[str] = None,
    ) -> Union["Parameter", ReActOutput]:
        r"""prompt_kwargs: additional prompt kwargs to either replace or add to the preset prompt kwargs."""
        # initialize step_history in both training and eval mode

        # set up the prompts
        prompt_kwargs = {
            **promt_kwargs,
            "input_str": input,
        }

        step_history: List[Union[StepOutput, Parameter]] = []
        if self.debug:
            printc(f"input_query: {input}", color="red")
        for i in range(self.max_steps):
            step = i + 1
            try:
                step_output = self._run_one_step(
                    step, prompt_kwargs, model_kwargs, id, step_history
                )
                if isinstance(step_output, Parameter):
                    step_output.data_id = id
                step_history.append(step_output)
                if self._check_last_step(step_history):
                    break
            except Exception as e:
                log.error(f"Error running step {step}: {e}")
                printc(f"Error running step {step}: {e}", color="red")
                raise e  # the only place to raise the error for debugging. In normal cases, the agent should not raise an error.

        answer = self._get_answer(step_history)
        if self.training:
            return answer
        # wrap the output
        output = ReActOutput(step_history=step_history, id=id, answer=answer)
        if self.debug:
            printc(f"answer: {output}", color="yellow")

        return output

    def _extra_repr(self) -> str:
        s = f"max_steps={self.max_steps}, add_llm_as_fallback={self.add_llm_as_fallback}, "
        return s


if __name__ == "__main__":
    from adalflow.components.model_client import OpenAIClient
    from adalflow.utils import setup_env
    from adalflow.core.func_tool import FunctionTool

    setup_env()

    class App(Component):
        def __init__(self):
            super().__init__()
            self.llm_tool = Generator(
                model_client=OpenAIClient(),
                model_kwargs={"model": "gpt-3.5-turbo"},
            )

            def llm_as_tool(input: str, id: Optional[str] = None) -> str:
                """Used as a calculator tool."""
                printc(f"llm_as_tool: {input}", color="yellow")

                return self.llm_tool(prompt_kwargs={"input_str": input}, id=id)

            self.react_agent = ReActAgent(
                tools=[FunctionTool(llm_as_tool, component=self.llm_tool)],
                max_steps=2,
                add_llm_as_fallback=False,
                model_client=OpenAIClient(),
                model_kwargs={"model": "gpt-3.5-turbo"},
            )

        def call(self, input: str, id: Optional[str] = None) -> Union[str, "Parameter"]:
            return self.react_agent(input, id=id)

        def forward(
            self, input: str, id: Optional[str] = None
        ) -> Union[str, "Parameter"]:
            return self.react_agent(input, id=id)

    # print(OutputParameter.__mro__)

    app = App()
    app.train()
    output = app("I want to multiply 3 and 4.", id="123")
    # print(output)
    printc(output, color="yellow")
    output.draw_graph()



r"""ModelClient is the protocol and base class for all models(either via APIs or local models) to communicate with components."""

from typing import Any, Dict, Optional


from adalflow.core.component import Component
from adalflow.core.types import (
    ModelType,
    EmbedderOutput,
    GeneratorOutput,
    CompletionUsage,
)


# TODO: global model registry for all available models in users' project.
class ModelClient(Component):
    __doc__ = r"""The protocol and abstract class for all models(either via APIs or local models) to communicate with components.

ModelClient is to separate the model API calls from the rest of the system,
making it a plug-and-play component that can be used in functional components like Generator and Embedder.


For a particular API provider, such as OpenAI, we will have a class that inherits from ModelClient.
It does four things:

(1) Initialize the client, including both sync and async.

(2) Convert the standard AdalFlow components inputs to the API-specific format.

(3) Call the API and parse the response.

(4) Handle API specific exceptions and errors to retry the call.

Check the subclasses in `components/model_client/` directory for the functional API clients we have.

    This interface is designed to bridge the gap between AdalFlow components inputs and model APIs.

    You can see examples of the subclasses in components/model_client/ directory.
    """

    def __init__(self, *args, **kwargs) -> None:
        r"""Ensure the subclasses will at least call self._init_sync_client() to initialize the sync client."""
        super().__init__()

        self.sync_client = None
        self.async_client = None

    def init_sync_client(self):
        raise NotImplementedError(
            f"{type(self).__name__} must implement _init_sync_client method"
        )

    def init_async_client(self):
        raise NotImplementedError(
            f"{type(self).__name__} must implement _init_async_client method"
        )

    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        r"""Subclass use this to call the API with the sync client.
        model_type: this decides which API, such as chat.completions or embeddings for OpenAI.
        api_kwargs: all the arguments that the API call needs, subclass should implement this method.

        Additionally in subclass you can implement the error handling and retry logic here. See OpenAIClient for example.
        """
        raise NotImplementedError(f"{type(self).__name__} must implement _call method")

    async def acall(
        self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED
    ):
        r"""Subclass use this to call the API with the async client."""
        pass

    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> Dict:
        r"""
        Bridge the Component's standard input and model_kwargs into API-specific format, the api_kwargs that will be used in _call and _acall methods.

        All types of models supported by this particular provider should be handled here.
        Args:
            input (Optional[Any], optional): input to the model. Defaults to None.
            model_kwargs (Dict): model kwargs
            model_type (ModelType): model type

        """
        raise NotImplementedError(
            f"{type(self).__name__} must implement _combine_input_and_model_kwargs method"
        )

    def parse_chat_completion(self, completion: Any) -> "GeneratorOutput":
        r"""Parse the chat completion to str."""
        raise NotImplementedError(
            f"{type(self).__name__} must implement parse_chat_completion method"
        )

    def track_completion_usage(self, *args, **kwargs) -> "CompletionUsage":
        r"""Track the chat completion usage. Use OpenAI standard API for tracking."""
        raise NotImplementedError(
            f"{type(self).__name__} must implement track_usage method"
        )

    def parse_embedding_response(self, response: Any) -> "EmbedderOutput":
        r"""Parse the embedding response to a structure AdalFlow components can understand."""
        raise NotImplementedError(
            f"{type(self).__name__} must implement parse_embedding_response method"
        )

    @staticmethod
    def _process_text(text: str) -> str:
        """This is specific to OpenAI API, as removing new lines could have better performance in the embedder"""
        text = text.replace("\n", " ")
        return text

    def _track_usage(self, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)



"""AzureOpenAI ModelClient integration."""

import os
from typing import (
    Dict,
    Sequence,
    Optional,
    List,
    Any,
    TypeVar,
    Callable,
    Generator,
    Union,
    Literal,
)
import re

import logging
import backoff

# optional import
from adalflow.utils.lazy_import import safe_import, OptionalPackages

import sys

openai = safe_import(OptionalPackages.OPENAI.value[0], OptionalPackages.OPENAI.value[1])
# Importing all Azure packages together
azure_modules = safe_import(
    OptionalPackages.AZURE.value[0],  # List of package names
    OptionalPackages.AZURE.value[1],  # Error message
)
# Manually add each module to sys.modules to make them available globally as if imported normally
azure_module_names = OptionalPackages.AZURE.value[0]
for name, module in zip(azure_module_names, azure_modules):
    sys.modules[name] = module

# Use the modules as if they were imported normally
from azure.identity import DefaultAzureCredential, get_bearer_token_provider

# from azure.core.credentials import AccessToken
from openai import AzureOpenAI, AsyncAzureOpenAI, Stream
from openai import (
    APITimeoutError,
    InternalServerError,
    RateLimitError,
    UnprocessableEntityError,
    BadRequestError,
)
from openai.types import (
    Completion,
    CreateEmbeddingResponse,
)
from openai.types.chat import ChatCompletionChunk, ChatCompletion

from adalflow.core.model_client import ModelClient
from adalflow.core.types import (
    ModelType,
    EmbedderOutput,
    TokenLogProb,
    CompletionUsage,
    GeneratorOutput,
)
from adalflow.components.model_client.utils import parse_embedding_response

log = logging.getLogger(__name__)
T = TypeVar("T")


__all__ = ["AzureAIClient"]

# TODO: this overlaps with openai client largely, might need to refactor to subclass openai client to simplify the code


# completion parsing functions and you can combine them into one singple chat completion parser
def get_first_message_content(completion: ChatCompletion) -> str:
    r"""When we only need the content of the first message.
    It is the default parser for chat completion."""
    return completion.choices[0].message.content


# def _get_chat_completion_usage(completion: ChatCompletion) -> OpenAICompletionUsage:
#     return completion.usage


def parse_stream_response(completion: ChatCompletionChunk) -> str:
    r"""Parse the response of the stream API."""
    return completion.choices[0].delta.content


def handle_streaming_response(generator: Stream[ChatCompletionChunk]):
    r"""Handle the streaming response."""
    for completion in generator:
        log.debug(f"Raw chunk completion: {completion}")
        parsed_content = parse_stream_response(completion)
        yield parsed_content


def get_all_messages_content(completion: ChatCompletion) -> List[str]:
    r"""When the n > 1, get all the messages content."""
    return [c.message.content for c in completion.choices]


def get_probabilities(completion: ChatCompletion) -> List[List[TokenLogProb]]:
    r"""Get the probabilities of each token in the completion."""
    log_probs = []
    for c in completion.choices:
        content = c.logprobs.content
        print(content)
        log_probs_for_choice = []
        for openai_token_logprob in content:
            token = openai_token_logprob.token
            logprob = openai_token_logprob.logprob
            log_probs_for_choice.append(TokenLogProb(token=token, logprob=logprob))
        log_probs.append(log_probs_for_choice)
    return log_probs


class AzureAIClient(ModelClient):
    __doc__ = r"""
    A client wrapper for interacting with Azure OpenAI's API.

    This class provides support for both embedding and chat completion API calls.
    Users can use this class to simplify their interactions with Azure OpenAI models
    through the `Embedder` and `Generator` components.

    **Initialization:**

    You can initialize the `AzureAIClient` with either an API key or Azure Active Directory (AAD) token
    authentication. It is recommended to set environment variables for sensitive data like API keys.

    Args:
        api_key (Optional[str]): Azure OpenAI API key. Default is None.
        api_version (Optional[str]): API version to use. Default is None.
        azure_endpoint (Optional[str]): Azure OpenAI endpoint URL. Default is None.
        credential (Optional[DefaultAzureCredential]): Azure AD credential for token-based authentication. Default is None.
        chat_completion_parser (Callable[[Completion], Any]): Function to parse chat completions. Default is `get_first_message_content`.
        input_type (Literal["text", "messages"]): Format for input, either "text" or "messages". Default is "text".

    **Setup Instructions:**

    - **Using API Key:**
      Set up the following environment variables:
      ```bash
      export AZURE_OPENAI_API_KEY="your_api_key"
      export AZURE_OPENAI_ENDPOINT="your_endpoint"
      export AZURE_OPENAI_VERSION="your_version"
      ```

    - **Using Azure AD Token:**
      Ensure you have configured Azure AD credentials. The `DefaultAzureCredential` will automatically use your configured credentials.

    **Example Usage:**

    .. code-block:: python

        from azure.identity import DefaultAzureCredential
        from your_module import AzureAIClient  # Adjust import based on your module name

        # Initialize with API key
        client = AzureAIClient(
            api_key="your_api_key",
            api_version="2023-05-15",
            azure_endpoint="https://your-endpoint.openai.azure.com/"
        )

        # Or initialize with Azure AD token
        client = AzureAIClient(
            api_version="2023-05-15",
            azure_endpoint="https://your-endpoint.openai.azure.com/",
            credential=DefaultAzureCredential()
        )

        # Example call to the chat completion API
        api_kwargs = {
            "model": "gpt-3.5-turbo",
            "messages": [{"role": "user", "content": "What is the meaning of life?"}],
            "stream": True
        }
        response = client.call(api_kwargs=api_kwargs, model_type=ModelType.LLM)

        for chunk in response:
            print(chunk)


    **Notes:**
    - Ensure that the API key or credentials are correctly set up and accessible to avoid authentication errors.
    - Use `chat_completion_parser` to define how to extract and handle the chat completion responses.
    - The `input_type` parameter determines how input is formatted for the API call.

    **References:**
    - [Azure OpenAI API Documentation](https://learn.microsoft.com/en-us/azure/ai-services/openai/overview)
    - [OpenAI API Documentation](https://platform.openai.com/docs/guides/text-generation)
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        api_version: Optional[str] = None,
        azure_endpoint: Optional[str] = None,
        credential: Optional[DefaultAzureCredential] = None,
        chat_completion_parser: Callable[[Completion], Any] = None,
        input_type: Literal["text", "messages"] = "text",
    ):
        r"""It is recommended to set the API_KEY into the  environment variable instead of passing it as an argument.


        Initializes the Azure OpenAI client with either API key or AAD token authentication.

        Args:
            api_key: Azure OpenAI API key.
            api_version: Azure OpenAI API version.
            azure_endpoint: Azure OpenAI endpoint.
            credential: Azure AD credential for token-based authentication.
            chat_completion_parser: Function to parse chat completions.
            input_type: Input format, either "text" or "messages".

        """
        super().__init__()

        # added api_type azure for azure Ai
        self.api_type = "azure"
        self._api_key = api_key
        self._apiversion = api_version
        self._azure_endpoint = azure_endpoint
        self._credential = credential
        self.sync_client = self.init_sync_client()
        self.async_client = None  # only initialize if the async call is called
        self.chat_completion_parser = (
            chat_completion_parser or get_first_message_content
        )
        self._input_type = input_type

    def init_sync_client(self):
        api_key = self._api_key or os.getenv("AZURE_OPENAI_API_KEY")
        azure_endpoint = self._azure_endpoint or os.getenv("AZURE_OPENAI_ENDPOINT")
        api_version = self._apiversion or os.getenv("AZURE_OPENAI_VERSION")
        # credential = self._credential or DefaultAzureCredential
        if not azure_endpoint:
            raise ValueError("Environment variable AZURE_OPENAI_ENDPOINT must be set")
        if not api_version:
            raise ValueError("Environment variable AZURE_OPENAI_VERSION must be set")

        if api_key:
            return AzureOpenAI(
                api_key=api_key, azure_endpoint=azure_endpoint, api_version=api_version
            )
        elif self._credential:
            # credential = DefaultAzureCredential()
            token_provider = get_bearer_token_provider(
                DefaultAzureCredential(), "https://cognitiveservices.azure.com/.default"
            )
            return AzureOpenAI(
                azure_ad_token_provider=token_provider,
                azure_endpoint=azure_endpoint,
                api_version=api_version,
            )
        else:
            raise ValueError(
                "Environment variable AZURE_OPENAI_API_KEY must be set or credential must be provided"
            )

    def init_async_client(self):
        api_key = self._api_key or os.getenv("AZURE_OPENAI_API_KEY")
        azure_endpoint = self._azure_endpoint or os.getenv("AZURE_OPENAI_ENDPOINT")
        api_version = self._apiversion or os.getenv("AZURE_OPENAI_VERSION")
        # credential = self._credential or DefaultAzureCredential()
        if not azure_endpoint:
            raise ValueError("Environment variable AZURE_OPENAI_ENDPOINT must be set")
        if not api_version:
            raise ValueError("Environment variable AZURE_OPENAI_VERSION must be set")

        if api_key:
            return AsyncAzureOpenAI(
                api_key=api_key, azure_endpoint=azure_endpoint, api_version=api_version
            )
        elif self._credential:
            # credential = DefaultAzureCredential()
            token_provider = get_bearer_token_provider(
                DefaultAzureCredential(), "https://cognitiveservices.azure.com/.default"
            )
            return AsyncAzureOpenAI(
                azure_ad_token_provider=token_provider,
                azure_endpoint=azure_endpoint,
                api_version=api_version,
            )
        else:
            raise ValueError(
                "Environment variable AZURE_OPENAI_API_KEY must be set or credential must be provided"
            )

    # def _parse_chat_completion(self, completion: ChatCompletion) -> "GeneratorOutput":
    #     # TODO: raw output it is better to save the whole completion as a source of truth instead of just the message
    #     try:
    #         data = self.chat_completion_parser(completion)
    #         usage = self.track_completion_usage(completion)
    #         return GeneratorOutput(
    #             data=data, error=None, raw_response=str(data), usage=usage
    #         )
    #     except Exception as e:
    #         log.error(f"Error parsing the completion: {e}")
    #         return GeneratorOutput(data=None, error=str(e), raw_response=completion)

    def parse_chat_completion(
        self,
        completion: Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]],
    ) -> "GeneratorOutput":
        """Parse the completion, and put it into the raw_response."""
        log.debug(f"completion: {completion}, parser: {self.chat_completion_parser}")
        try:
            data = self.chat_completion_parser(completion)
            usage = self.track_completion_usage(completion)
            return GeneratorOutput(
                data=None, error=None, raw_response=data, usage=usage
            )
        except Exception as e:
            log.error(f"Error parsing the completion: {e}")
            return GeneratorOutput(data=None, error=str(e), raw_response=completion)

    def track_completion_usage(
        self,
        completion: Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]],
    ) -> CompletionUsage:
        if isinstance(completion, ChatCompletion):
            usage: CompletionUsage = CompletionUsage(
                completion_tokens=completion.usage.completion_tokens,
                prompt_tokens=completion.usage.prompt_tokens,
                total_tokens=completion.usage.total_tokens,
            )
            return usage
        else:
            raise NotImplementedError(
                "streaming completion usage tracking is not implemented"
            )

    def parse_embedding_response(
        self, response: CreateEmbeddingResponse
    ) -> EmbedderOutput:
        r"""Parse the embedding response to a structure LightRAG components can understand.

        Should be called in ``Embedder``.
        """
        try:
            return parse_embedding_response(response)
        except Exception as e:
            log.error(f"Error parsing the embedding response: {e}")
            return EmbedderOutput(data=[], error=str(e), raw_response=response)

    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> Dict:
        r"""
        Specify the API input type and output api_kwargs that will be used in _call and _acall methods.
        Convert the Component's standard input, and system_input(chat model) and model_kwargs into API-specific format
        """

        final_model_kwargs = model_kwargs.copy()
        if model_type == ModelType.EMBEDDER:
            if isinstance(input, str):
                input = [input]
            # convert input to input
            if not isinstance(input, Sequence):
                raise TypeError("input must be a sequence of text")
            final_model_kwargs["input"] = input
        elif model_type == ModelType.LLM:
            # convert input to messages
            messages: List[Dict[str, str]] = []

            if self._input_type == "messages":
                system_start_tag = "<START_OF_SYSTEM_PROMPT>"
                system_end_tag = "<END_OF_SYSTEM_PROMPT>"
                user_start_tag = "<START_OF_USER_PROMPT>"
                user_end_tag = "<END_OF_USER_PROMPT>"
                pattern = f"{system_start_tag}(.*?){system_end_tag}{user_start_tag}(.*?){user_end_tag}"
                # Compile the regular expression
                regex = re.compile(pattern)
                # Match the pattern
                match = regex.search(input)
                system_prompt, input_str = None, None

                if match:
                    system_prompt = match.group(1)
                    input_str = match.group(2)

                else:
                    print("No match found.")
                if system_prompt and input_str:
                    messages.append({"role": "system", "content": system_prompt})
                    messages.append({"role": "user", "content": input_str})
            if len(messages) == 0:
                messages.append({"role": "system", "content": input})
            final_model_kwargs["messages"] = messages
        else:
            raise ValueError(f"model_type {model_type} is not supported")
        return final_model_kwargs

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
            BadRequestError,
        ),
        max_time=5,
    )
    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        """
        kwargs is the combined input and model_kwargs.  Support streaming call.
        """
        log.info(f"api_kwargs: {api_kwargs}")
        if model_type == ModelType.EMBEDDER:
            return self.sync_client.embeddings.create(**api_kwargs)
        elif model_type == ModelType.LLM:
            if "stream" in api_kwargs and api_kwargs.get("stream", False):
                log.debug("streaming call")
                self.chat_completion_parser = handle_streaming_response
                return self.sync_client.chat.completions.create(**api_kwargs)
            return self.sync_client.chat.completions.create(**api_kwargs)
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
            BadRequestError,
        ),
        max_time=5,
    )
    async def acall(
        self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED
    ):
        """
        kwargs is the combined input and model_kwargs
        """
        if self.async_client is None:
            self.async_client = self.init_async_client()
        if model_type == ModelType.EMBEDDER:
            return await self.async_client.embeddings.create(**api_kwargs)
        elif model_type == ModelType.LLM:
            return await self.async_client.chat.completions.create(**api_kwargs)
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @classmethod
    def from_dict(cls: type[T], data: Dict[str, Any]) -> T:
        obj = super().from_dict(data)
        # recreate the existing clients
        obj.sync_client = obj.init_sync_client()
        obj.async_client = obj.init_async_client()
        return obj

    def to_dict(self) -> Dict[str, Any]:
        r"""Convert the component to a dictionary."""
        # TODO: not exclude but save yes or no for recreating the clients
        exclude = [
            "sync_client",
            "async_client",
        ]  # unserializable object
        output = super().to_dict(exclude=exclude)
        return output


# if __name__ == "__main__":
#     from adalflow.core import Generator
#     from adalflow.utils import setup_env, get_logger

#     log = get_logger(level="DEBUG")

#     setup_env()
#     prompt_kwargs = {"input_str": "What is the meaning of life?"}

#     gen = Generator(
#         model_client=OpenAIClient(),
#         model_kwargs={"model": "gpt-3.5-turbo", "stream": True},
#     )
#     gen_response = gen(prompt_kwargs)
#     print(f"gen_response: {gen_response}")

#     for genout in gen_response.data:
#         print(f"genout: {genout}")



"""Google GenAI ModelClient integration."""

import os
from typing import Dict, Sequence, Optional, Any
import backoff
import logging


from adalflow.core.model_client import ModelClient
from adalflow.core.types import CompletionUsage, ModelType, GeneratorOutput

from adalflow.utils.lazy_import import safe_import, OptionalPackages

# optional import
google = safe_import(
    OptionalPackages.GOOGLE_GENERATIVEAI.value[0],
    OptionalPackages.GOOGLE_GENERATIVEAI.value[1],
)
import google.generativeai as genai
from google.api_core.exceptions import (
    InternalServerError,
    BadRequest,
    GoogleAPICallError,
)
from google.generativeai.types import GenerateContentResponse

log = logging.getLogger(__name__)


class GoogleGenAIClient(ModelClient):
    __doc__ = r"""A component wrapper for the Google GenAI API client.

    Visit https://cloud.google.com/vertex-ai/generative-ai/docs/model-reference/inference for more api details.

    Info: 8/1/2024
    Tested: gemini-1.0-pro, gemini-1.5-pro-latest

    .. code-block:: python

        class UsageMetadata(proto.Message):

            prompt_token_count: int = proto.Field(
                proto.INT32,
                number=1,
            )
            cached_content_token_count: int = proto.Field(
                proto.INT32,
                number=4,
            )
            candidates_token_count: int = proto.Field(
                proto.INT32,
                number=2,
            )
            total_token_count: int = proto.Field(
                proto.INT32,
                number=3,
            )
    """

    def __init__(self, api_key: Optional[str] = None):
        r"""It is recommended to set the GOOGLE_API_KEY environment variable instead of passing it as an argument."""
        super().__init__()
        self._api_key = api_key
        self.sync_client = self.init_sync_client()
        self.async_client = None  # only initialize if the async call is called
        self.tested_llm_models = ["gemini-1.0-pro", "gemini-1.5-pro-latest"]

    def init_sync_client(self):
        api_key = self._api_key or os.getenv("GOOGLE_API_KEY")
        if not api_key:
            raise ValueError("Environment variable GOOGLE_API_KEY must be set")
        genai.configure(api_key=api_key)
        return genai

    def parse_chat_completion(
        self, completion: GenerateContentResponse
    ) -> "GeneratorOutput":
        """
        Parse the completion to a structure your sytem standarizes. (here is str)
        """
        log.debug(f"completion: {completion}")
        try:
            data = completion.text
            usage = self.track_completion_usage(completion)
            return GeneratorOutput(data=None, usage=usage, raw_response=data)
        except Exception as e:
            log.error(f"Error parsing completion: {e}")
            return GeneratorOutput(
                data=None, error=str(e), raw_response=str(completion)
            )

    def track_completion_usage(
        self, completion: GenerateContentResponse
    ) -> CompletionUsage:
        return CompletionUsage(
            completion_tokens=completion.usage_metadata.candidates_token_count,
            prompt_tokens=completion.usage_metadata.prompt_token_count,
            total_tokens=completion.usage_metadata.total_token_count,
        )

    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> Dict:
        r"""
        Specify the API input type and output api_kwargs that will be used in _call and _acall methods.
        Convert the Component's standard input, and system_input(chat model) and model_kwargs into API-specific format
        """
        final_model_kwargs = model_kwargs.copy()
        if model_type == ModelType.EMBEDDER:
            if isinstance(input, str):
                input = [input]
            # convert input to input
            if not isinstance(input, Sequence):
                raise TypeError("input must be a sequence of text")
            final_model_kwargs["input"] = input
        elif model_type == ModelType.LLM:

            final_model_kwargs["prompt"] = input
        else:
            raise ValueError(f"model_type {model_type} is not supported")
        return final_model_kwargs

    @backoff.on_exception(
        backoff.expo,
        (
            InternalServerError,
            BadRequest,
            GoogleAPICallError,
        ),
        max_time=5,
    )
    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        """
        kwargs is the combined input and model_kwargs
        """

        if model_type == ModelType.LLM:
            # remove model from api_kwargs
            model = api_kwargs.pop("model")
            prompt = api_kwargs.pop("prompt")

            assert model != "", "model must be specified"
            assert prompt != "", "prompt must be specified"

            config = genai.GenerationConfig(**api_kwargs)
            llm = genai.GenerativeModel(model_name=model, generation_config=config)
            return llm.generate_content(contents=prompt)
        else:
            raise ValueError(f"model_type {model_type} is not supported")


if __name__ == "__main__":
    from adalflow.utils import setup_env

    setup_env()
    google_client = GoogleGenAIClient()



from adalflow.database.sqlalchemy.sqlachemy_manager import DatabaseManager

#  we need to import the DocumentModel class to create the tables
from adalflow.database.sqlalchemy.model import DocumentModel  # noqa


if __name__ == "__main__":
    db_name = "vector_db"
    postgres_url = f"postgresql://postgres:password@localhost:5432/{db_name}"

    db_manager = DatabaseManager(postgres_url)
    db_manager.create_tables()
    print(db_manager.list_table_schemas())
    db_manager.close()



"""Components here are used for data processing/transformation."""

from .text_splitter import TextSplitter
from .data_components import ToEmbeddings, RetrieverOutputToContextStr
from adalflow.utils.registry import EntityMapping


__all__ = ["TextSplitter", "ToEmbeddings", "RetrieverOutputToContextStr"]

for name in __all__:
    EntityMapping.register(name, globals()[name])



import csv
import json
import os
import uuid
from typing import Literal
import subprocess
from adalflow.utils.data import Dataset
from adalflow.datasets.types import Example

from adalflow.utils.file_io import save_csv
from adalflow.datasets.utils import prepare_dataset_path


# TODO: here users clean adalflow created files
class BigBenchHard(Dataset):
    __doc__ = """Big Bench Hard dataset for object counting task.

    You can find the task name from the following link:
    https://github.com/suzgunmirac/BIG-Bench-Hard/tree/main/bbh


    Data will be saved to ~/.adalflow/cache_datasets/BBH_object_counting/{split}.csv
    if root is not specified.

    Size for each split:
    - train: 50 examples
    - val: 100 examples
    - test: 100 examples

    Args:
        task_name (str): The name of the task. "{task_name}" is the task name in the dataset.
        root (str, optional): Root directory of the dataset to save the data. Defaults to ~/.adalflow/cache_datasets/task_name.
        split (str, optional): The dataset split, supports ``"train"`` (default), ``"val"`` and ``"test"``.
    """

    def __init__(
        self,
        task_name: Literal["object_counting"] = "object_counting",
        root: str = None,
        split: Literal["train", "val", "test"] = "train",
        *args,
        **kwargs,
    ):

        if split not in ["train", "val", "test"]:
            raise ValueError("Split must be one of 'train', 'val', 'test'")

        # if root is None:
        #     root = get_adalflow_default_root_path()
        #     print(f"Saving dataset to {root}")
        self.root = root
        self.split = split

        self.task_name = task_name
        data_path = prepare_dataset_path(self.root, self.task_name)
        self._check_or_download_dataset(data_path, split)

        self.data = []
        split_csv_path = os.path.join(data_path, f"{split}.csv")
        with open(split_csv_path, newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                self.data.append(
                    Example(question=row["x"], answer=row["y"], id=row["id"])
                )  # dont use a tuple, use a dict {"x": ..., "y": ...}

    def _check_or_download_dataset(self, data_path: str = None, split: str = "train"):

        if data_path is None:
            raise ValueError("data_path must be specified")
        json_path = os.path.join(data_path, f"{self.task_name}.json")
        split_csv_path = os.path.join(data_path, f"{split}.csv")
        if os.path.exists(split_csv_path):
            return

        print(f"Downloading dataset to {json_path}")
        try:
            # Use subprocess and capture the return code
            result = subprocess.call(
                [
                    "wget",
                    f"https://raw.githubusercontent.com/suzgunmirac/BIG-Bench-Hard/main/bbh/{self.task_name}.json",
                    "-O",
                    json_path,
                ]
            )

            # Check if wget failed (non-zero exit code)
            if result != 0:
                raise ValueError(
                    f"Failed to download dataset for task '{self.task_name}'.\n"
                    "Please verify the task name (the JSON file name) by checking the following link:\n"
                    "https://github.com/suzgunmirac/BIG-Bench-Hard/tree/main/bbh"
                )

            # Check if the file is non-empty
            if not os.path.exists(json_path) or os.path.getsize(json_path) == 0:
                raise ValueError(
                    f"Downloaded file is empty. Please check the task name '{self.task_name}' or network issues."
                )

        except Exception as e:
            raise ValueError(
                f"Either network issues or an incorrect task name: '{self.task_name}'.\n"
                "Please verify the task name (the JSON file name) by checking the following link:\n"
                "https://github.com/suzgunmirac/BIG-Bench-Hard/tree/main/bbh"
            ) from e

        with open(json_path) as json_file:
            data = json.load(json_file)

        examples = data["examples"]

        # NOTE: better to shuffle the examples before splitting.
        # We do this splitting in order to be consistent with text-grad paper.

        train_examples = [
            {"x": ex["input"], "y": ex["target"], "id": str(uuid.uuid4())}
            for ex in examples[:50]
        ]
        val_examples = [
            {"x": ex["input"], "y": ex["target"], "id": str(uuid.uuid4())}
            for ex in examples[50:150]
        ]
        test_examples = [
            {"x": ex["input"], "y": ex["target"], "id": str(uuid.uuid4())}
            for ex in examples[150:]
        ]
        # ensure the

        for split, examples in zip(
            ["train", "val", "test"], [train_examples, val_examples, test_examples]
        ):
            target_path = os.path.join(data_path, f"{split}.csv")
            print(f"Saving {split} split to {target_path}")
            save_csv(examples, f=target_path, fieldnames=["x", "y", "id"])

    def __getitem__(self, index) -> Example:
        return self.data[index]

    def __len__(self):
        return len(self.data)

    @staticmethod
    def get_default_task_instruction():
        _task_description = "You will answer a reasoning question. Think step by step. The last line of your response should be of the following format: 'Answer: $VALUE' where VALUE is a numerical value."
        return _task_description


if __name__ == "__main__":
    from adalflow.datasets.big_bench_hard import BigBenchHard

    dataset = BigBenchHard(task_name="object_counting", split="test")
    print(dataset[0:10])
    print(len(dataset))
    print(dataset.get_default_task_instruction())



"""
This logger file provides easy configurability of the root and named loggers, along with a color print function for console output.

NOTE: If you are on Windows versions prior to Windows 10, the Command Prompt and PowerShell do not support ANSI color codes. You would need to use third-party libraries like colorama to enable ANSI color support.
Please add the following colorama setting at the beginning of your code:

Installation:

.. code-block:: bash

    pip install colorama

Initialization:

.. code-block:: python

    import colorama
    colorama.init(autoreset=True)
"""

import logging
import sys
from typing import List, Tuple, Optional, Literal
import inspect
import os
from datetime import datetime
from adalflow.utils.global_config import get_adalflow_default_root_path

LOG_LEVELS = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}

# color map, from color string to colorama.Fore colors, easier to use, reduce friction
COLOR_MAP = {
    "black": "\x1b[30m",
    "blue": "\x1b[34m",
    "cyan": "\x1b[36m",
    "green": "\x1b[32m",
    "magenta": "\x1b[35m",
    "red": "\x1b[31m",
    "white": "\x1b[37m",
    "yellow": "\x1b[33m",
}

log = logging.getLogger(__name__)


def _get_log_config(
    level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
    filepath: str = "./logs/app.log",
    enable_console: bool = True,
    enable_file: bool = True,
) -> Tuple[int, List[logging.Handler]]:
    r"""Helper function to get the default log configuration.

    We config logging with the following default settings:
    1. Enable both console and file output.
    2. Set up the default format: "%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d:%(funcName)s] - %(message)s".
    The format is: time, log level, filename(where the log is from), line number, function name, message.
    3. Set up the default date format: "%Y-%m-%d %H:%M:%S".

    Args:
        level (str): Log level. Defaults to "INFO". Options: "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL".
        filepath (str): Name of the output log file. Defaults to "./logs/app.log".
        enable_console (bool): Control the console output. Defaults to True.
        enable_file (bool): Control the file output. Defaults to True.

    Returns:
        Tuple[int, List[logging.Handler]]: The logging level and the list of handlers.
    """

    def get_level(level: str) -> int:
        """Return the logging level constant based on a string."""
        return LOG_LEVELS.get(level.upper(), logging.INFO)

    # 2. Config the default format and style
    format = "%(asctime)s - %(module)s - %(levelname)s - [%(filename)s:%(lineno)d:%(funcName)s] - %(message)s"
    formatter = logging.Formatter(format, datefmt="%Y-%m-%d %H:%M:%S")

    # Set up the handler
    handlers: List[logging.Handler] = []
    if enable_console:
        handlers.append(logging.StreamHandler(sys.stdout))
    if enable_file:
        file_path: str = os.path.dirname(filepath)
        os.makedirs(file_path, exist_ok=True)
        handler = logging.FileHandler(filename=filepath, mode="a")
        handlers.append(handler)

    for h in handlers:
        h.setFormatter(formatter)

    return get_level(level), handlers


def get_logger(
    name: Optional[str] = None,
    level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
    save_dir: Optional[str] = None,
    filename: Optional[str] = None,
    enable_console: bool = True,
    enable_file: bool = True,
) -> logging.Logger:
    """Get or configure a logger, including both the root logger and named logger.

    Args:
        name (Optional[str]): Name of the logger. Defaults to None which means root logger.
        level (str): Log level. Defaults to "INFO". Options: "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL".
        save_dir (Optional[str]): Directory to save log files. Defaults to "./logs".
        filename (Optional[str]): Name of the output log file. Defaults to "lib.log" for root logger and "{name}.log" for named logger.
        enable_console (bool): Control the console output. Defaults to True.
        enable_file (bool): Control the file output. Defaults to True.

    Returns:
        logging.Logger: The logger with the specified configuration.

    Example:

    1. Get the root logger and have it log to both console and file (~/.adalflow/logs/lib.log):

    .. code-block:: python

        from adalflow.utils.logger import get_logger

        root_logger = get_logger(enable_console=True, enable_file=True)

    2. Get a named logger and have it log to both console and file (./logs/app.log):

    .. code-block:: python

        logger = get_logger(name="app", enable_console=True, enable_file=True)

    3. Use two loggers, one for the library and one for the application:

    .. code-block:: python

        root_logger = enable_library_logging(level="DEBUG", enable_console=True, enable_file=True)
        logger = get_logger(name="app", level="DEBUG", enable_console=True, enable_file=True)
        logger.info("This is an application log, and it will be saved in app.log separately from the library log at lib.log.")
    """
    # save_dir = save_dir or "./logs"
    save_dir = save_dir or get_adalflow_default_root_path() + "/logs"

    os.makedirs(save_dir, exist_ok=True)
    if name is None:
        filename = filename or "lib.log"
    else:
        filename = filename or f"{name}.log"
    filepath = os.path.join(save_dir, filename)

    config = _get_log_config(
        level=level,
        filepath=filepath,
        enable_console=enable_console,
        enable_file=enable_file,
    )
    logger = logging.getLogger(name)
    logger.setLevel(config[0])
    logger.handlers = []  # Clear existing handlers
    for handler in config[1]:
        logger.addHandler(handler)
    if name:
        logger.propagate = False  # Do not propagate to the root logger
    return logger


def get_current_script_and_line() -> Tuple[str, str, int]:
    """Get the function name, script name, and line of where the print is called.

    Returns:
        function_name (str): the name of the function that is calling.
        script_name (str): the name of the script that is calling.
        line_number (str): the line number of the code that is calling.
    """
    caller_frame = inspect.stack()[2]
    file_path, line_number, function_name = (
        caller_frame.filename,
        caller_frame.lineno,
        caller_frame.function,
    )
    script_name = os.path.basename(file_path)

    return function_name, script_name, line_number


def printc(
    text: str,
    color: Literal[
        "black", "blue", "cyan", "green", "magenta", "red", "white", "yellow"
    ] = "cyan",
):
    """Color enhanced print function with logger-like format.

    LightRAG's customized print with colored text, position of code block the print is set, and current timestamp.

    Args:
        text (str): Text to be printed.
        color (str): Color of the text. Defaults to "cyan". Options:
        'black', 'blue', 'cyan', 'green', 'magenta', 'red', 'white', 'yellow'. Defaults to "cyan".

    Example:
        .. code-block:: python

            from utils.logger import colored_print
            printc("hello", color="green")
            printc("hello")
    """
    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    function_name, script_name, line_number = get_current_script_and_line()
    color_code = COLOR_MAP.get(color, COLOR_MAP["cyan"])
    print(
        f"{color_code}{timestamp} - [{script_name}:{line_number}:{function_name}] - {text}\033[0m"
    )
    # \033[0m means reset, not impacting the next print texts



"""Lazy import a module and class."""

from typing import List, Union
import importlib
import logging
from types import ModuleType

from enum import Enum

log = logging.getLogger(__name__)


class OptionalPackages(Enum):
    __doc__ = r"""Enum for optional packages that can be used in the library.

    The package name and error message are defined for each optional package as a tuple.

    The value of the tuple:
    - The package name (str): The package name to import. Follows the right syntax: such as import azure.identity but the package itself is azure-identity.
      Support a list of package names for related packages. This will be importing a list of packages while safe_import is used.
    - The error message (str): The message to display if the package is not found.

    Example of using multiple related packages:

    .. code-block:: python

        from adalflow.utils.lazy_import import safe_import, OptionalPackages
        import sys

        azure_modules = safe_import(
        OptionalPackages.AZURE.value[0],  # List of package names
        OptionalPackages.AZURE.value[1],  # Error message
        )
        # Manually add each module to sys.modules to make them available globally as if imported normally
        azure_module_names = OptionalPackages.AZURE.value[0]
        for name, module in zip(azure_module_names, azure_modules):
            sys.modules[name] = module

        # Use the modules as if they were imported normally
        from azure.identity import DefaultAzureCredential, get_bearer_token_provider
    """
    # model sdk
    GROQ = ("groq", "Please install groq with: pip install groq")
    OPENAI = ("openai", "Please install openai with: pip install openai")
    ANTHROPIC = ("anthropic", "Please install anthropic with: pip install anthropic")
    GOOGLE_GENERATIVEAI = (
        "google.generativeai",
        "Please install google-generativeai with: pip install google-generativeai",
    )
    TRANSFORMERS = (
        "transformers",
        "Please install transformers with: pip install transformers",
    )
    COHERE = ("cohere", "Please install cohere with: pip install cohere")
    OLLAMA = ("ollama", "Please install ollama with: pip install ollama")
    # AWS
    BOTO3 = (
        ["boto3", "botocore"],
        "Please install boto3 and botocore with: pip install boto3 botocore",
    )
    # modeling library
    TORCH = ("torch", "Please install torch with: pip install torch")

    # Grouping all Azure-related packages under one entry
    AZURE = (
        [
            "azure.identity",
            "azure.core",
            # "azure.ai-formrecognizer",
            # "azure.ai-textanalytics",
        ],
        "Please install Azure packages with: pip install azure-identity azure-core azure-ai-formrecognizer azure-ai-textanalytics",
    )
    # search library
    FAISS = (
        "faiss",
        "Please install faiss with: pip install faiss-cpu (or faiss if you use GPU)",
    )

    LANCEDB = (
        "lancedb",
        "Please install lancedb with: pip install lancedb .",
    )

    # db library
    SQLALCHEMY = (
        "sqlalchemy",
        "Please install sqlalchemy with: pip install sqlalchemy",
    )
    PGVECTOR = (
        "pgvector",
        "Please install pgvector with: pip install pgvector",
    )
    DATASETS = (
        "datasets",
        "Please install datasets with: pip install datasets",
    )
    QDRANT = (
        "qdrant-client",
        "Please install qdrant-client with: pip install qdrant-client",
    )

    def __init__(self, package_name, error_message):
        self.package_name = package_name
        self.error_message = error_message


class LazyImport:
    __doc__ = r"""Lazy import a module and class.

    It is a proxy. The class/func will be created only when the class is instantiated or called.

    .. note::

       Do not subclass a lazy imported class.

    Mainly internal library use to import optional packages only when needed.

    Args:
        import_path (str): The import path of the module and class, eg. "adalflow.components.model_client.openai_client.OpenAIClient".
        optional_package (OptionalPackages): The optional package to import, it helps define the package name and error message.
    """

    def __init__(
        self, import_path: str, optional_package: OptionalPackages, *args, **kwargs
    ):
        if args or kwargs:
            raise TypeError(
                "LazyImport does not support subclassing or additional arguments. "
                "Import the class directly from its specific module instead. For example, "
                "from adalflow.components.model_client.cohere_client import CohereAPIClient"
                "instead of using the lazy import with: from adalflow.components.model_client import CohereAPIClient"
            )
        self.import_path = import_path
        self.optional_package = optional_package
        self.module = None
        self.class_ = None

    def load_class(self):
        log.debug(f"Loading class: {self.import_path}")
        if self.module is None:
            try:
                components = self.import_path.split(".")
                module_path = ".".join(components[:-1])
                class_name = components[-1]
            except Exception as e:
                log.error(f"Error parsing import path: {e}")
            try:
                self.module = importlib.import_module(module_path)
                self.class_ = getattr(self.module, class_name)
                return self.class_
            except ImportError:
                log.info(f"Optional module not installed: {self.import_path}")
                raise ImportError(f"{self.optional_package.error_message}")

    def __getattr__(self, name):
        self.load_class()
        return getattr(self.class_, name)

    def __call__(self, *args, **kwargs):
        """Create the class instance."""
        self.load_class()
        log.debug(f"Creating class instance: {self.class_}")
        # normal class initialization
        return self.class_(*args, **kwargs)


def safe_import(
    module_names: Union[str, List[str]], install_message: str
) -> Union[ModuleType, List[ModuleType]]:
    """Safely import a module and raise an ImportError with the install message if the module is not found.

    Handles importing of multiple related packages.

    Args:
        module_names (list or str): The package name(s) to import.
        install_message (str): The message to display if import fails.

    Returns:
        ModuleType: The imported module.

    Raises:
        ImportError: If any of the packages are not found.


    Example:

    1. Tests

    .. code-block:: python

        try:
            numpy = safe_import("numpy", "Please install numpy with: pip install numpy")
            print(numpy.__version__)
        except ImportError as e:
            print(e)

    When numpy is not installed, it will raise an ImportError with the install message.
    When numpy is installed, it will print the numpy version.

    2. Use it to delay the import of optional packages in the library.

    .. code-block:: python

        from adalflow.utils.lazy_import import safe_import, OptionalPackages

        numpy = safe_import(OptionalPackages.NUMPY.value[0], OptionalPackages.NUMPY.value[1])

    """
    if isinstance(module_names, str):
        module_names = [module_names]

    return_modules = []
    for module_name in module_names:
        try:
            return_modules.append(importlib.import_module(module_name))
        except ImportError:
            raise ImportError(f"{install_message}")

    return return_modules[0] if len(return_modules) == 1 else return_modules



"""Inspired by TextGrad: Automatic differentiation via "text" """

from typing import TYPE_CHECKING, Callable


if TYPE_CHECKING:
    from adalflow.core.generator import BackwardEngine


class BackwardContext:
    """
    Represents a context for backward computation.

    :param backward_fn: The backward function to be called during backward computation.
    :type backward_fn: callable
    :param args: Variable length argument list to be passed to the backward function.
    :param kwargs: Arbitrary keyword arguments to be passed to the backward function.

    :ivar backward_fn: The backward function to be called during backward computation.
    :vartype backward_fn: callable
    :ivar fn_name: The fully qualified name of the backward function.
    :vartype fn_name: str
    :ivar args: Variable length argument list to be passed to the backward function.
    :ivar kwargs: Arbitrary keyword arguments to be passed to the backward function.

    :method __call__(backward_engine: EngineLM) -> Any:
        Calls the backward function with the given backward engine and returns the result.
    :method __repr__() -> str:
        Returns a string representation of the BackwardContext object.
    """

    def __init__(
        self,
        disable_backward_engine: bool,
        backward_fn: Callable,
        backward_engine: "BackwardEngine" = None,
        *args,
        **kwargs,
    ):
        self.backward_fn = backward_fn
        self.disable_backward_engine = disable_backward_engine
        self.backward_engine = backward_engine
        self.fn_name = f"{backward_fn.__module__}.{backward_fn.__qualname__}"
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        if self.backward_engine is None:
            return self.backward_fn(*self.args, **self.kwargs)

        return self.backward_fn(
            *self.args,
            **self.kwargs,
            backward_engine=self.backward_engine,
            disable_backward_engine=self.disable_backward_engine,
        )

    def __repr__(self):
        return f"{self.fn_name}"



from .optimizer import Optimizer
from .sampler import RandomSampler, ClassSampler, Sampler
from .parameter import Parameter, OutputParameter
from .function import BackwardContext
from .few_shot.bootstrap_optimizer import BootstrapFewShot
from .text_grad.tgd_optimizer import TGDOptimizer
from .text_grad.text_loss_with_eval_fn import EvalFnToTextLoss
from .text_grad.llm_text_loss import LLMAsTextLoss
from .trainer.trainer import Trainer
from .trainer.adal import AdalComponent
from adalflow.utils.registry import EntityMapping
from .optimizer import DemoOptimizer, TextOptimizer
from .gradient import Gradient, GradientContext


__all__ = [
    "Optimizer",
    "RandomSampler",
    "ClassSampler",
    "Sampler",
    "Parameter",
    "OutputParameter",
    "BackwardContext",
    "BootstrapFewShot",
    "TGDOptimizer",
    "EvalFnToTextLoss",
    "LLMAsTextLoss",
    "Trainer",
    "AdalComponent",
    "Optimizer",
    "DemoOptimizer",
    "TextOptimizer",
    "Gradient",
    "GradientContext",
]

for name in __all__:
    EntityMapping.register(name, globals()[name])



from .react import DEFAULT_REACT_AGENT_SYSTEM_PROMPT, ReActAgent
from adalflow.utils.registry import EntityMapping

__all__ = [
    "ReActAgent",
    "DEFAULT_REACT_AGENT_SYSTEM_PROMPT",
]

for name in __all__:
    EntityMapping.register(name, globals()[name])



import uuid
from dataclasses import dataclass, field
from typing import Dict
from adalflow.core.base_data_class import DataClass


@dataclass
class BaseData(DataClass):
    __doc__ = """A common dataclass for representing examples in a dataset."""
    id: str = field(
        metadata={"desc": "The unique identifier of the example", "type": "id"},
        default=str(uuid.uuid4()),
    )


@dataclass
class Example(DataClass):
    __doc__ = """A common dataclass for representing examples in a dataset."""
    id: str = field(
        metadata={"desc": "The unique identifier of the example"},
        default=str(uuid.uuid4()),
    )
    question: str = field(
        metadata={"desc": "The question to be answered"}, default=None
    )
    answer: str = field(metadata={"desc": "The answer to the question"}, default=None)


@dataclass
class HotPotQAData(Example):
    __doc__ = """A dataclass for representing examples in the HotPotQA dataset."""
    gold_titles: set = field(
        metadata={"desc": "The set of titles that support the answer"},
        default=None,
    )
    context: Dict[str, object] = field(
        metadata={"desc": "The context of the question"},
        default=None,
    )

    __input_fields__ = ["question"]
    __output_fields__ = ["answer"]

    # @staticmethod
    # def from_dict(d: Dict[str, Any]) -> "HotPotQAData":
    #     # Preprocess gold_titles
    #     if "gold_titles" in d and isinstance(d["gold_titles"], str):
    #         try:
    #             d["gold_titles"] = json.loads(d["gold_titles"])
    #         except json.JSONDecodeError:
    #             # Replace single quotes with double quotes
    #             fixed_str = d["gold_titles"].replace("'", '"')
    #             d["gold_titles"] = set(json.loads(fixed_str))

    #     # Preprocess context
    #     if "context" in d and isinstance(d["context"], str):
    #         try:
    #             d["context"] = json.loads(d["context"])
    #         except json.JSONDecodeError:
    #             fixed_str = d["context"].replace("'", '"')
    #             d["context"] = json.loads(fixed_str)

    #     return HotPotQAData(**d)


@dataclass
class TrecData(BaseData):
    __doc__ = """A dataclass for representing examples in the TREC dataset."""
    question: str = field(
        metadata={"desc": "The question to be classified"},
        default=None,
    )
    class_name: str = field(
        metadata={"desc": "One of {ABBR, ENTY, DESC, HUM, LOC, NUM}"},
        default=None,
    )
    class_index: int = field(
        metadata={"desc": "The class label, in range [0, 5]"},
        default=-1,
    )

    __input_fields__ = ["question"]  # follow this order too.
    __output_fields__ = ["class_name", "class_index"]


if __name__ == "__main__":
    # test the hotpotqa data
    data = HotPotQAData(
        question="What is the capital of France?",
        answer="Paris",
        gold_titles=set(["Paris", "France"]),
        context={"Paris": "The capital of France"},
    )

    data_dict = data.to_dict()
    print("data_dict", data_dict)
    data = HotPotQAData.from_dict(data_dict)
    print("data", data)

    from adalflow.utils.file_io import save_json, load_json

    # save json
    save_json(data_dict, f="task.json")
    # load json
    data_dict_loaded = load_json(f="task.json")

    print("data_dict_loaded", data_dict_loaded)

    # restore the data
    data_restored = HotPotQAData.from_dict(data_dict_loaded)
    print("data_restored", data_restored)



from .serialization import (
    default,
    serialize,
    deserialize,
)
from .file_io import (
    save_json,
    load_json,
    save_pickle,
    load_pickle,
    save,
    load,
    load_jsonl,
    append_to_jsonl,
    write_list_to_jsonl,
)
from .logger import printc, get_logger
from .registry import EntityMapping
from .config import new_components_from_config, new_component
from .lazy_import import LazyImport, OptionalPackages, safe_import
from .setup_env import setup_env
from .data import DataLoader, Dataset, Subset
from .global_config import get_adalflow_default_root_path
from .cache import CachedEngine


__all__ = [
    "save",
    "load",
    "printc",
    "get_logger",
    "EntityMapping",
    "new_components_from_config",
    "LazyImport",
    "OptionalPackages",
    "new_component",
    "default",
    "serialize",
    "deserialize",
    "save_json",
    "load_json",
    "save_pickle",
    "load_pickle",
    "load_jsonl",
    "append_to_jsonl",
    "write_list_to_jsonl",
    "safe_import",
    "setup_env",
    "DataLoader",
    "Dataset",
    "Subset",
    "get_adalflow_default_root_path",
    "CachedEngine",
]



"""AdalComponent provides an interface to compose different parts, from eval_fn, train_step, loss_step, optimizers, backward engine, teacher generator, etc to work with Trainer."""

from typing import Any, Callable, Dict, List, Optional, Tuple, TYPE_CHECKING
import concurrent
from tqdm import tqdm
import numpy as np
import warnings
import logging as log

if TYPE_CHECKING:
    from adalflow.core.model_client import ModelClient
    from adalflow.core.generator import Generator, BackwardEngine, BackwardPassSetup
    from adalflow.optim.parameter import Parameter

from adalflow.core.component import Component
from adalflow.core.types import GeneratorOutput
from adalflow.optim.optimizer import Optimizer
from adalflow.optim.loss_component import LossComponent
from adalflow.optim.types import PromptData
from adalflow.eval.base import EvaluationResult
from adalflow.optim.grad_component import GradComponent
from adalflow.utils import printc

from adalflow.optim.optimizer import DemoOptimizer, TextOptimizer


log = log.getLogger(__name__)


class AdalComponent(Component):
    __doc__ = """Define a train, eval, and test step for a task pipeline.

    This serves the following purposes:
    1. Organize all parts for training a task pipeline in one place.
    2. Help with debugging and testing before the actual training.
    3. Adds multi-threading support for training and evaluation.

    It has no need on call, forward, bicall, or __call__, so we need to overwrite the base ones.
    """

    task: Component
    # evaluator: Optional[BaseEvaluator]
    eval_fn: Optional[Callable]  # final eval score
    loss_fn: Optional[LossComponent]
    loss_eval_fn: Optional[Callable]  # loss eval score and training subset eval fn
    backward_engine: Optional["BackwardEngine"]
    _demo_optimizers: Optional[List[DemoOptimizer]]
    _text_optimizers: Optional[List[TextOptimizer]]

    def __init__(
        self,
        task: Component,
        # evaluator: Optional[BaseEvaluator] = None,
        eval_fn: Optional[Callable] = None,
        loss_eval_fn: Optional[Callable] = None,
        loss_fn: Optional[LossComponent] = None,
        backward_engine: Optional["BackwardEngine"] = None,
        backward_engine_model_config: Optional[Dict] = None,
        teacher_model_config: Optional[Dict] = None,
        text_optimizer_model_config: Optional[Dict] = None,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.task = task
        # self.evaluator = evaluator
        self.eval_fn = eval_fn
        self.loss_eval_fn = loss_eval_fn
        self.loss_fn = loss_fn
        self.backward_engine = backward_engine
        if backward_engine and not isinstance(backward_engine, "BackwardEngine"):
            raise ValueError(
                f"backward_engine is not a BackwardEngine: {backward_engine}"
            )
        self.backward_engine_model_config = backward_engine_model_config
        self.teacher_model_config = teacher_model_config
        self.text_optimizer_model_config = text_optimizer_model_config
        self._demo_optimizers = None
        self._text_optimizers = None

    def _set_param_values(self, prompts: List[PromptData]):
        r"""Set the parameters for the task. Used to resume from ckpt."""

        params_dict = {p.name: p for p in prompts}

        for name, param in self.task.named_parameters():
            if name in params_dict:
                param.update_value(params_dict[name].data)

    def _get_param_values(self) -> List[PromptData]:
        r"""Get the current values of the parameters."""
        return [
            PromptData(p.id, p.name, p.data, p.requires_opt)
            for p in self.task.parameters()
            if p.requires_opt
        ]

    def prepare_task(self, sample: Any, *args, **kwargs) -> Tuple[Callable, Dict]:
        r"""Tell Trainer how to call the task in both training and inference mode.

        Return a task call and kwargs for one training sample.

        If you just need to eval, ensure the Callable has the inference mode.
        If you need to also train, ensure the Callable has the training mode which returns a Parameter and mainly
        call forward for all subcomponents within the task.

        Example:

        .. code-block:: python

            def prepare_task(self, sample: Any, *args, **kwargs) -> Tuple[Callable, Dict]:
                return self.task, {"x": sample.x}
        """

        raise NotImplementedError("prepare_task method is not implemented")

    def prepare_loss(
        self, sample: Any, y_pred: "Parameter", *args, **kwargs
    ) -> Tuple[Callable, Dict]:
        r"""Tell Trainer how to calculate the loss in the training mode.

        Return a loss call and kwargs for one loss sample.

        Need to ensure y_pred is a Parameter, and the real input to use
        for y_gt and y_pred is `eval_input`.
        Make sure it is setup.

        Example:

        .. code-block:: python

            # "y" and "y_gt" are arguments needed
            #by the eval_fn inside of the loss_fn if it is a EvalFnToTextLoss

            def prepare_loss(self, sample: Example, pred: adal.Parameter) -> Dict:
                # prepare gt parameter
                y_gt = adal.Parameter(
                    name="y_gt",
                    data=sample.answer,
                    eval_input=sample.answer,
                    requires_opt=False,
                )

                # pred's full_response is the output of the task pipeline which is GeneratorOutput
                pred.eval_input = pred.full_response.data
                return self.loss_fn, {"kwargs": {"y": y_gt, "y_pred": pred}}
        """
        raise NotImplementedError("prepare_loss method is not implemented")

    # TODO: Support multiple eval_fn with different metrics. using a dict[str, (Callable, Dict)] to store them.
    def prepare_eval(self, sample: Any, y_pred: Any, *args, **kwargs) -> float:
        r"""Tell Trainer how to eval in inference mode.
        Return the eval_fn and kwargs for one evaluation sample.

        Ensure the eval_fn is a callable that takes the predicted output and the ground truth output.
        Ensure the kwargs are setup correctly.

        """
        raise NotImplementedError("prepare_eval method is not implemented")

    def prepare_loss_eval(self, sample: Any, y_pred: Any, *args, **kwargs) -> float:
        r"""Tell Trainer how to eval in inference mode.
        Return the eval_fn and kwargs for one evaluation sample.

        Ensure the eval_fn is a callable that takes the predicted output and the ground truth output.
        Ensure the kwargs are setup correctly.

        """
        raise NotImplementedError("prepare_loss_eval method is not implemented")

    # def configure_optimizers(self, *args, **kwargs) -> Optimizer:
    #     r"""Note: When you use text optimizor, ensure you call `configure_backward_engine_engine` too."""
    #     raise NotImplementedError("configure_optimizers method is not implemented")

    def configure_optimizers(self, *args, **text_optimizer_kwargs) -> List[Optimizer]:
        r"""Note: When you use text optimizor, ensure you call `configure_backward_engine_engine` too."""
        if self._demo_optimizers is None:
            self._demo_optimizers = self.configure_demo_optimizer_helper()
        if self._text_optimizers is None:
            if not self.text_optimizer_model_config:
                raise ValueError("Text optimizer model config is not configured.")
            if not self.text_optimizer_model_config.get("model_client"):
                raise ValueError("Model client is not configured.")
            if not self.text_optimizer_model_config.get("model_kwargs"):
                raise ValueError("Model kwargs is not configured.")

            self._text_optimizers = self.configure_text_optimizer_helper(
                **self.text_optimizer_model_config, **text_optimizer_kwargs
            )
        return self._demo_optimizers + self._text_optimizers

    def configure_backward_engine(self, *args, **kwargs):
        r"""Configure a backward engine for all GradComponent in the task for bootstrapping examples."""
        # check if backward engine is already configured
        if self.backward_engine:
            log.warning("Backward engine is already configured.")
        if not self.backward_engine_model_config:
            raise ValueError("Backward engine model config is not configured.")
        if not self.backward_engine_model_config.get("model_client"):
            raise ValueError("Model client is not configured.")
        if not self.backward_engine_model_config.get("model_kwargs"):
            raise ValueError("Model kwargs is not configured.")
        self.configure_backward_engine_helper(
            model_client=self.backward_engine_model_config["model_client"],
            model_kwargs=self.backward_engine_model_config["model_kwargs"],
            backward_pass_setup=kwargs.get("backward_pass_setup", None),
        )

    def disable_backward_engine(self):
        r"""Disable the backward engine for all GradComponent in the task.
        No more gradients generation."""
        self.disable_backward_engine_helper()

    # def configure_backward_engine(self, *args, **kwargs):
    #     raise NotImplementedError("configure_backward_engine method is not implemented")

    def evaluate_samples(
        self,
        samples: Any,
        y_preds: List,
        metadata: Optional[Dict[str, Any]] = None,
        num_workers: int = 2,
        use_loss_eval_fn: bool = False,
    ) -> EvaluationResult:
        r"""Evaluate predictions against the ground truth samples.
        Run evaluation on samples using parallel processing. Utilizes ``prepare_eval`` defined by the user.

        Metadata is used for storing context that you can find from generator input.

        Args:
            samples (Any): The input samples to evaluate.
            y_preds (List): The predicted outputs corresponding to each sample.
            metadata (Optional[Dict[str, Any]]): Optional metadata dictionary.
            num_workers (int): Number of worker threads for parallel processing.

        Returns:
            EvaluationResult: An object containing the average score and per-item scores.
        """
        if use_loss_eval_fn and not self.loss_eval_fn:
            raise ValueError("Loss eval function is not configured.")

        from adalflow.optim.parameter import Parameter

        if not isinstance(y_preds, list) or len(y_preds) == 0:
            raise ValueError(f"y_preds is not a list or empty: {y_preds}")

        y_pred_0 = y_preds[0]
        if isinstance(y_pred_0, Parameter):
            raise ValueError(f"y_pred_0 should not be a Parameter: {y_pred_0}")

        acc_list = [None] * len(samples)  # Initialize accuracy list to hold results

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:

            # 1. submit all the tasks
            futures = {}

            for i, (sample, y_pred) in enumerate(zip(samples, y_preds)):

                if metadata is None:
                    if not use_loss_eval_fn:
                        eval_fn, kwargs = self.prepare_eval(sample, y_pred)
                    else:
                        eval_fn, kwargs = self.prepare_loss_eval(sample, y_pred)
                    future = executor.submit(eval_fn, **kwargs)
                    # future = executor.submit(self.evaluate_one_sample, sample, y_pred)
                else:
                    if not use_loss_eval_fn:
                        eval_fn, kwargs = self.prepare_eval(
                            sample, y_pred, metadata=metadata
                        )
                    else:

                        eval_fn, kwargs = self.prepare_eval(
                            sample, y_pred, metadata=metadata
                        )
                    future = executor.submit(eval_fn, **kwargs)
                    # future = executor.submit(
                    #     self.evaluate_one_sample, sample, y_pred, metadata=metadata
                    # )
                futures[future] = i

            # 2. collect the results, update the progress bar
            # Initialize progress bar once outside the loop
            progress_bar = tqdm(
                total=len(samples), desc="Evaluating", position=0, leave=True
            )

            for future in concurrent.futures.as_completed(futures):
                try:
                    i = futures[future]
                    acc_list[i] = (
                        future.result()
                    )  # Place the result in the correct position
                    progress_bar.update(
                        1
                    )  # Update progress bar after each result is collected
                except Exception as e:

                    progress_bar.close()
                    raise ValueError(f"Exception in task {i}: {e}")

        avg_score = float(np.mean(np.array(acc_list)))
        return EvaluationResult(avg_score=avg_score, per_item_scores=acc_list)

    # def evaluate_samples(
    #     self, samples: Any, y_preds: List, metadata: Optional[Dict[str, Any]] = None
    # ) -> EvaluationResult:
    #     r"""Run evaluation on samples. Use ``evaluate_one_sample`` defined by the user.

    #     Metadata is used for storing context that you can find from generator input.

    #     Note:
    #         ensure it supports both Tuple(batch) and a list of any type (fits for datasets).
    #     """
    #     from adalflow.optim.parameter import Parameter

    #     if not isinstance(y_preds, list) or len(y_preds) == 0:
    #         raise ValueError(f"y_preds is not a list or empty: {y_preds}")
    #     y_pred_0 = y_preds[0]
    #     if isinstance(y_pred_0, Parameter):
    #         raise ValueError(f"y_pred_0 should not be a Parameter: {y_pred_0}")
    #     if metadata is None:
    #         acc_list = [
    #             self.evaluate_one_sample(sample, y_pred)
    #             for sample, y_pred in zip(samples, y_preds)
    #         ]
    #     else:
    #         acc_list = [
    #             self.evaluate_one_sample(sample, y_pred, metadata=metadata)
    #             for sample, y_pred in zip(samples, y_preds)
    #         ]
    #     avg_score = float(np.mean(np.array(acc_list)))
    #     return EvaluationResult(avg_score=avg_score, per_item_scores=acc_list)

    def _train_step(
        self,
        batch,
        batch_idx,
        num_workers: int = 2,
    ):
        r"""Applies to both train and eval mode.

        If you require self.task.train() to be called before training, you can override this method as:

        .. code-block:: python

            def train_step(self, batch, batch_idx, num_workers: int = 2) -> List:
                self.task.train()
                return super().train_step(batch, batch_idx, num_workers)
        """
        from adalflow.optim.parameter import Parameter

        self.task.train()

        y_preds = [None] * len(batch)
        samples = [None] * len(batch)

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
            futures = []
            tqdm_loader = tqdm(batch, total=len(batch), desc="Loading Data")
            for i, sample in enumerate(tqdm_loader):
                task_call, kwargs = self.prepare_task(sample)
                future = executor.submit(task_call, **kwargs)
                futures.append((future, i, sample))  # preserve the order of the samples

            tqdm_loader = tqdm(
                total=len(futures),
                position=0,
                desc="Training",
            )
            for future, i, sample in futures:
                y_pred = future.result()
                y_preds[i] = y_pred  # Place the prediction in the correct position
                samples[i] = sample  # Keep the sample order aligned
                if not isinstance(y_pred, Parameter):
                    raise ValueError(f"y_pred_{i} is not a Parameter, {y_pred}")

                if hasattr(y_pred, "full_response") and isinstance(
                    y_pred.full_response, GeneratorOutput
                ):
                    if y_pred.full_response.id is not None:
                        y_pred_sample_id = y_pred.full_response.id
                        assert (
                            y_pred_sample_id == sample.id
                        ), f"ID mismatch: {y_pred_sample_id} != {sample.id}, type: {type(y_pred)}"

                tqdm_loader.update(1)  # Update the progress bar

        return y_preds

    def pred_step(
        self,
        batch,
        batch_idx,
        num_workers: int = 2,
        running_eval: bool = False,
        min_score: Optional[float] = None,
        use_loss_eval_fn: bool = False,
    ) -> Tuple[List["Parameter"], List, Dict[int, float]]:
        r"""Applies to only the eval mode.

        Args:
            batch (Any): The input batch to predict.
            batch_idx (int): The index of the batch.
            num_workers (int): Number of worker threads for parallel processing.
            running_eval: bool = False,
        Returns:
            Tuple[List["Parameter"], List, Dict[int, float]]: The predicted outputs, the samples, and the scores.
        """
        from adalflow.optim.parameter import Parameter

        self.task.eval()

        y_preds = [None] * len(batch)
        samples = [None] * len(batch)
        completed_indices = set()
        index_to_score = {}  # for running evaluation

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
            # 1. submit all the tasks
            futures = []
            tqdm_loader = tqdm(batch, total=len(batch), desc="Loading Data")
            for i, sample in enumerate(tqdm_loader):
                task_call, kwargs = self.prepare_task(sample)
                future = executor.submit(task_call, **kwargs)
                futures.append((future, i, sample))  # preserve the order of the samples

            # 2. predict the results, update the progress bar
            tqdm_loader = tqdm(
                total=len(futures),
                position=0,
                desc=f"Prediting step: {batch_idx}",
            )
            for future, i, sample in futures:
                y_pred = future.result()
                y_preds[i] = y_pred  # Place the prediction in the correct position
                samples[i] = sample  # Keep the sample order aligned
                # check the ordering

                if isinstance(y_pred, Parameter):
                    raise ValueError(f"y_pred_{i} is a Parameter, {y_pred}")

                assert (
                    y_pred.id == sample.id
                ), f"ID mismatch: {y_pred.id} != {sample.id}, type: {type(y_pred)}"

                completed_indices.add(i)  # Mark this index as completed

                if running_eval and not isinstance(y_pred, Parameter):
                    # evaluate one sample

                    if not use_loss_eval_fn:
                        eval_fn, kwargs = self.prepare_eval(sample, y_pred)
                    else:
                        eval_fn, kwargs = self.prepare_loss_eval(sample, y_pred)
                    score = eval_fn(**kwargs)
                    index_to_score[i] = score
                    eval_score = np.mean(list(index_to_score.values())).item()

                    # for y_pred, sample in zip(completed_y_preds, completed_samples):
                    #     print(f"y_pred: {y_pred.data}, sample: {sample.answer}")
                    # for y_pred, sample in zip(completed_y_preds, completed_samples):
                    #     if y_pred.id != sample.id:
                    #         raise ValueError(
                    #             f"ID mismatch: {y_pred.id} != {sample.id}, type: {type(y_pred)}"
                    #         )
                    #     print(f"y_pred: {y_pred.data}, sample: {sample.answer}")

                    remaining_samples = len(batch) - len(completed_indices)
                    max_score = (
                        eval_score * len(completed_indices) + remaining_samples
                    ) / len(batch)

                    if min_score is not None and max_score < min_score:
                        break

                    tqdm_loader.set_description(
                        f"Predicting: step({batch_idx}): {round(eval_score,4)} across {len(completed_indices)} samples, Max potential: {round(max_score,4)}"
                    )
                else:
                    tqdm_loader.set_description(f"Predicting: step({batch_idx})")

                tqdm_loader.update(1)  # Update the progress bar

        sorted_indices = sorted(completed_indices)
        completed_y_preds = [y_preds[idx] for idx in sorted_indices]
        completed_samples = [samples[idx] for idx in sorted_indices]

        return completed_y_preds, completed_samples, index_to_score

    def train_step(self, batch, batch_idx, num_workers: int = 2) -> List:
        r"""Run a training step and return the predicted outputs.
        Likely a list of Parameters."""
        self.task.train()
        y_preds = self._train_step(batch, batch_idx, num_workers)
        for i, y_pred in enumerate(y_preds):
            try:
                y_pred.name += f"y_pred_{i}"
            except AttributeError:
                raise ValueError(f"y_pred_{i} is not a Parameter, {y_pred}")
        return y_preds

    def validate_condition(self, steps: int, total_steps: int) -> bool:
        r"""In default, trainer will validate at every step."""
        return True

    def validation_step(
        self,
        batch,
        batch_idx,
        num_workers: int = 2,
        minimum_score: Optional[float] = None,
        use_loss_eval_fn: bool = False,
    ) -> EvaluationResult:
        r"""
        Args:
            batch (Any): The input batch to validate, can be a whole dataset
            batch_idx (int): The index of the batch. or current_step
            num_workers (int): Number of worker threads for parallel processing.
            minimum_score (Optional[float]): The max potential score needs to be larger than this to continue evaluating.

        Evaluate a batch or the validate dataset by setting the batch=val_dataset.
        Uses self.eval_fn to evaluate the samples.
        If you require self.task.eval() to be called before validation, you can override this method as:

        .. code-block:: python

            def validation_step(self, batch, batch_idx, num_workers: int = 2) -> List:
                self.task.eval()
                return super().validation_step(batch, batch_idx, num_workers)
        """
        # TODO: let use decide which mode to be
        eval_fn = self.eval_fn
        if use_loss_eval_fn:
            eval_fn = self.loss_eval_fn
            if not eval_fn:
                raise ValueError("Loss eval function is not configured.")

        self.task.eval()
        self.task.use_teacher(mode=False)  # ensure the teacher is not used
        try:
            completed_y_preds, completed_samples, index_to_score = self.pred_step(
                batch,
                batch_idx,
                num_workers,
                running_eval=True,
                min_score=minimum_score,
                use_loss_eval_fn=use_loss_eval_fn,
            )
        except Exception as e:
            raise ValueError(f"Error in validation step: {e}")

        if index_to_score:
            # compute score from index_to_score

            avg_score = np.mean(list(index_to_score.values())).item()
            acc_list = [None] * len(index_to_score)
            for i, score in index_to_score.items():
                acc_list[i] = score
            acc_list = list(index_to_score.values())
            eval_results = EvaluationResult(
                avg_score=avg_score, per_item_scores=acc_list
            )
        else:
            try:

                eval_results = self.evaluate_samples(
                    samples=completed_samples,
                    y_preds=completed_y_preds,
                    num_workers=num_workers,
                    use_loss_eval_fn=use_loss_eval_fn,
                )
            except Exception as e:
                raise ValueError(f"Error in evaluation: {e}")
        return eval_results

    def loss_step(
        self, batch, y_preds: List["Parameter"], batch_idx, num_workers: int = 2
    ) -> List["Parameter"]:
        r"""Calculate the loss for the batch."""
        from adalflow.optim.parameter import Parameter

        losses = [None] * len(batch)
        with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
            futures = []
            tqdm_loader = tqdm(
                zip(batch, y_preds), total=len(batch), desc="Loading Data"
            )
            for i, (sample, y_pred) in enumerate(tqdm_loader):
                loss_forward, kwargs = self.prepare_loss(sample, y_pred)
                future = executor.submit(loss_forward, **kwargs)
                futures.append((future, i, sample))
            tqdm_loader = tqdm(
                total=len(futures),
                position=0,
                desc="Calculating Loss",
            )
            for future, i, sample in futures:

                loss = future.result()
                if not isinstance(loss, Parameter):
                    raise ValueError(f"Loss is not a Parameter: {loss}")
                losses[i] = loss
                tqdm_loader.update(1)
        return losses

    # def configure_teacher_generator(self):
    #     r"""Configure a teach generator for all generators in the task for bootstrapping examples.

    #     You can call `configure_teacher_generator_helper` to easily configure it by passing the model_client and model_kwargs.
    #     """
    #     raise NotImplementedError(
    #         "configure_teacher_generator method is not implemented"
    #     )

    # use default implementation
    def configure_teacher_generator(self):
        r"""Configure a teach generator for all generators in the task for bootstrapping examples.

        You can call `configure_teacher_generator_helper` to easily configure it by passing the model_client and model_kwargs.
        """
        if not self.teacher_model_config:
            raise ValueError("Teacher model config is not configured.")
        if not self.teacher_model_config.get("model_client"):
            raise ValueError("Model client is not configured.")
        if not self.teacher_model_config.get("model_kwargs"):
            raise ValueError("Model kwargs is not configured.")
        print("Configuring teacher generator.")
        self.configure_teacher_generator_helper(**self.teacher_model_config)

    def configure_teacher_generator_helper(
        self,
        model_client: "ModelClient",
        model_kwargs: Dict[str, Any],
        template: Optional[str] = None,
    ):
        r"""Configure a teach generator for all generators in the task for bootstrapping examples."""
        from adalflow.core.generator import create_teacher_generator

        all_generators = self._find_all_generators()
        for _, generator in all_generators:
            teacher_generator = create_teacher_generator(
                student=generator,
                model_client=model_client,
                model_kwargs=model_kwargs,
                template=template,
            )
            print(f"Configuring teacher generator for {teacher_generator}")
            generator.set_teacher_generator(teacher_generator)
        print("Teacher generator configured.")

    def disable_backward_engine_helper(self):
        r"""Disable the backward engine for all gradcomponents in the task."""
        all_grads = self._find_all_grad_components()
        for _, grad in all_grads:
            if hasattr(grad, "disable_backward_engine") and callable(
                getattr(grad, "disable_backward_engine", None)
            ):
                grad.disable_backward_engine()
        print("Backward engine disabled for GradComponents")

        if not self.loss_fn:
            raise ValueError("Loss function is not configured.")

        # configure it for loss_fn
        if self.loss_fn:
            self.loss_fn.disable_backward_engine()

    def configure_backward_engine_helper(
        self,
        model_client: "ModelClient",
        model_kwargs: Dict[str, Any],
        template: Optional[str] = None,
        backward_pass_setup: Optional["BackwardPassSetup"] = None,
    ):
        r"""Configure a backward engine for all generators in the task for bootstrapping examples."""
        from adalflow.core.generator import BackwardEngine

        self.backward_engine = BackwardEngine(
            model_client=model_client,
            model_kwargs=model_kwargs,
            template=template,
        )
        if backward_pass_setup is not None:
            self.backward_engine.update_default_backward_pass_setup(backward_pass_setup)

        # set all generator's backward engine

        all_grads = self._find_all_grad_components()
        for _, grad in all_grads:
            if hasattr(grad, "set_backward_engine") and callable(
                getattr(grad, "set_backward_engine", None)
            ):
                grad.set_backward_engine(self.backward_engine)
        print("Backward engine configured for GradComponents")

        if not self.loss_fn:
            raise ValueError("Loss function is not configured.")

        # configure it for loss_fn
        if self.loss_fn:
            self.loss_fn.set_backward_engine(self.backward_engine)

    def configure_callbacks(
        self, save_dir: Optional[str] = "traces", *args, **kwargs
    ) -> List[str]:
        """In default we config the failure generator callback. User can overwrite this method to add more callbacks."""
        from adalflow.utils.global_config import get_adalflow_default_root_path
        import os

        if not save_dir:
            save_dir = "traces"
            save_dir = os.path.join(get_adalflow_default_root_path(), save_dir)
        log.debug(f"Saving traces to {save_dir}")
        return self._auto_generator_callbacks(save_dir)

    def run_one_task_sample(self, sample: Any) -> Any:
        r"""Run one training sample. Used for debugging and testing."""
        training = self.task.training
        # test training
        self.task.train()
        task_call, kwargs = self.prepare_task(sample)
        output = task_call(**kwargs)
        if not isinstance(output, Parameter):
            warnings.warn(f"Output is not a Parameter in training mode: {output}")
        # eval mode
        self.task.eval()
        task_call, kwargs = self.prepare_task(sample)
        output = task_call(**kwargs)
        if isinstance(output, Parameter):
            warnings.warn(f"Output is a Parameter in evaluation mode: {output}")
        # reset training
        self.task.train(training)

    def run_one_loss_sample(self, sample: Any, y_pred: Any) -> Any:
        r"""Run one loss sample. Used for debugging and testing."""
        loss_call, kwargs = self.prepare_loss(sample, y_pred)
        return loss_call(**kwargs)

    def _find_all_generators(self) -> List[Tuple[str, "Generator"]]:
        r"""Find all generators automatically from the task."""
        from adalflow.core import Generator

        all_generators: List[Tuple[str, Generator]] = []
        for name, comp in self.task.named_components():
            if isinstance(comp, Generator):
                all_generators.append((name, comp))
        log.debug(f"all_generators: {all_generators}")
        return all_generators

    def _find_all_grad_components(self) -> List[Tuple[str, GradComponent]]:
        r"""Find all generators automatically from the task."""
        # from adalflow.core import Generator

        all_grads: List[Tuple[str, GradComponent]] = []
        for name, comp in self.task.named_components():
            if isinstance(comp, GradComponent) or isinstance(comp, GradComponent):
                all_grads.append((name, comp))
        log.debug(f"all_grads: {all_grads}")
        return all_grads

    def _auto_generator_callbacks(self, save_dir: str = "traces") -> List[str]:
        r"""Automatically generate callbacks."""
        from adalflow.core.types import GeneratorOutput
        from adalflow.tracing.generator_call_logger import (
            GeneratorCallLogger,
        )
        from functools import partial

        all_generators = self._find_all_generators()

        log.debug(f"all_generators: {all_generators}")

        def _on_completion_callback(
            output: GeneratorOutput,
            input: Dict[str, Any],
            prompt_kwargs: Dict[str, Any],
            model_kwargs: Dict[str, Any],
            logger_call: Callable,
        ):
            r"""Log the generator output."""
            logger_call(
                output=output,
                input=input,
                prompt_kwargs=prompt_kwargs,
                model_kwargs=model_kwargs,
            )

        # Register the callback for each generator

        file_paths = []
        call_logger = GeneratorCallLogger(save_dir=save_dir)
        for name, generator in all_generators:

            # call_logger.reset()
            call_logger.register_generator(name)
            logger_call = partial(call_logger.log_call, name)
            generator.register_callback(
                "on_complete", partial(_on_completion_callback, logger_call=logger_call)
            )
            file_path = call_logger.get_log_location(name)
            file_paths.append(file_path)
            log.debug(f"Registered callback for {name}, file path: {file_path}")
        return file_paths

    def configure_demo_optimizer_helper(self) -> List[DemoOptimizer]:
        r"""One demo optimizer can handle multiple demo parameters.
        But the demo optimizer will only have one dataset (trainset) configured by the Trainer.

        If users want to use different trainset for different demo optimizer,
        they can configure it by themselves.
        """
        from adalflow.optim.few_shot.bootstrap_optimizer import BootstrapFewShot
        from adalflow.optim.parameter import ParameterType

        parameters = []
        for name, param in self.task.named_parameters():
            param.name = name
            if not param.param_type == ParameterType.DEMOS:
                continue
            parameters.append(param)
        if len(parameters) == 0:
            print("No demo parameters found.")
            return []
        do = BootstrapFewShot(params=parameters)
        return [do]

    def configure_text_optimizer_helper(
        self, model_client: "ModelClient", model_kwargs: Dict[str, Any], **kwargs
    ) -> List[TextOptimizer]:
        r"""Text optimizer hands prompt parameter type. One text optimizer can handle multiple text parameters."""
        from adalflow.optim.text_grad.tgd_optimizer import TGDOptimizer
        from adalflow.optim.parameter import ParameterType

        parameters = []
        for name, param in self.task.named_parameters():
            param.name = name
            if not param.param_type == ParameterType.PROMPT:
                continue
            parameters.append(param)
        if not parameters:
            return []

        to = TGDOptimizer(
            params=parameters,
            model_client=model_client,
            model_kwargs=model_kwargs,
            **kwargs,
        )

        printc(
            f"Text optimizer configured for {len(parameters)} parameters. names: { [(p.name, p.data) for p in parameters] }"
        )
        return [to]

    def _extra_repr(self):
        s = f"eval_fn: {self.eval_fn.__name__}, backward_engine: {self.backward_engine}, "
        s += f"backward_engine_model_config: {self.backward_engine_model_config}, teacher_model_config: {self.teacher_model_config}, text_optimizer_model_config: {self.text_optimizer_model_config}"
        return s

    def __call__(self, *args, **kwargs):
        pass

    def bicall(self, *args, **kwargs):
        pass

    def call(self, *args, **kwargs):
        pass

    def forward(self, *args, **kwargs):
        pass



"""A simple callback manager for tracing Geneator calls"""

from typing import Callable, Literal


class CallbackManager:
    def __init__(self):
        self.callbacks = {
            "on_success": [],
            "on_failure": [],
            "on_complete": [],  # You can define more event types as needed.
        }

    def register_callback(
        self,
        event_type: Literal["on_success", "on_failure", "on_complete"],
        callback: Callable,
    ):
        """Register a callback for a specific event type."""
        if event_type in self.callbacks:
            self.callbacks[event_type].append(callback)
        else:
            raise ValueError(f"Unsupported event type: {event_type}")

    def trigger_callbacks(
        self,
        event_type: Literal["on_success", "on_failure", "on_complete"],
        *args,
        **kwargs,
    ):
        """Invoke all callbacks for a given event type."""
        if event_type in self.callbacks:
            for callback in self.callbacks[event_type]:
                callback(*args, **kwargs)
        else:
            raise ValueError(f"Unsupported event type: {event_type}")



from .outputs import (
    YamlOutputParser,
    JsonOutputParser,
    ListOutputParser,
    YAML_OUTPUT_FORMAT,
    JSON_OUTPUT_FORMAT,
    LIST_OUTPUT_FORMAT,
)
from .dataclass_parser import DataClassParser

__all__ = [
    "YamlOutputParser",
    "JsonOutputParser",
    "ListOutputParser",
    "YAML_OUTPUT_FORMAT",
    "JSON_OUTPUT_FORMAT",
    "LIST_OUTPUT_FORMAT",
    "DataClassParser",
]



"""Anthropic ModelClient integration."""

import os
from typing import Dict, Optional, Any, Callable
import backoff
import logging


from adalflow.core.model_client import ModelClient
from adalflow.core.types import ModelType, CompletionUsage, GeneratorOutput

# optional import
from adalflow.utils.lazy_import import safe_import, OptionalPackages

anthropic = safe_import(
    OptionalPackages.ANTHROPIC.value[0], OptionalPackages.ANTHROPIC.value[1]
)

# import anthropic
from anthropic import (
    RateLimitError,
    APITimeoutError,
    InternalServerError,
    UnprocessableEntityError,
    BadRequestError,
)
from anthropic.types import Message, Usage

log = logging.getLogger(__name__)


def get_first_message_content(completion: Message) -> str:
    r"""When we only need the content of the first message.
    It is the default parser for chat completion."""
    return completion.content[0].text


__all__ = ["AnthropicAPIClient", "get_first_message_content"]


# NOTE: using customize parser might make the new_component more complex when we have to handle a callable
class AnthropicAPIClient(ModelClient):
    __doc__ = r"""A component wrapper for the Anthropic API client.

    Visit https://docs.anthropic.com/en/docs/intro-to-claude for more api details.

    Note:

    As antropic API needs users to set max_tokens, we set up a default value of 512 for the max_tokens.
    You can override this value by passing the max_tokens in the model_kwargs.

    Reference: 8/1/2024
    - https://docs.anthropic.com/en/docs/about-claude/models
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        chat_completion_parser: Callable[[Message], Any] = None,
    ):
        r"""It is recommended to set the ANTHROPIC_API_KEY environment variable instead of passing it as an argument."""
        super().__init__()
        self._api_key = api_key
        self.sync_client = self.init_sync_client()
        self.async_client = None  # only initialize if the async call is called
        self.tested_llm_models = ["claude-3-opus-20240229"]
        self.chat_completion_parser = (
            chat_completion_parser or get_first_message_content
        )
        self.default_max_tokens = 512

    def init_sync_client(self):
        api_key = self._api_key or os.getenv("ANTHROPIC_API_KEY")
        if not api_key:
            raise ValueError("Environment variable ANTHROPIC_API_KEY must be set")
        return anthropic.Anthropic(api_key=api_key)

    def init_async_client(self):
        api_key = self._api_key or os.getenv("ANTHROPIC_API_KEY")
        if not api_key:
            raise ValueError("Environment variable ANTHROPIC_API_KEY must be set")
        return anthropic.AsyncAnthropic(api_key=api_key)

    def parse_chat_completion(self, completion: Message) -> GeneratorOutput:
        log.debug(f"completion: {completion}")
        try:
            data = completion.content[0].text
            usage = self.track_completion_usage(completion)
            return GeneratorOutput(data=None, usage=usage, raw_response=data)
        except Exception as e:
            log.error(f"Error parsing completion: {e}")
            return GeneratorOutput(
                data=None, error=str(e), raw_response=str(completion)
            )

    def track_completion_usage(self, completion: Message) -> CompletionUsage:
        r"""Track the completion usage."""
        usage: Usage = completion.usage
        return CompletionUsage(
            completion_tokens=usage.output_tokens,
            prompt_tokens=usage.input_tokens,
            total_tokens=usage.output_tokens + usage.input_tokens,
        )

    # TODO: potentially use <SYS></SYS> to separate the system and user messages. This requires user to follow it. If it is not found, then we will only use user message.
    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> dict:
        r"""Anthropic API messages separates the system and the user messages.

        As we focus on one prompt, we have to use the user message as the input.

        api: https://docs.anthropic.com/en/api/messages
        """
        api_kwargs = model_kwargs.copy()
        if model_type == ModelType.LLM:
            api_kwargs["messages"] = [
                {"role": "user", "content": input},
            ]
            if "max_tokens" not in api_kwargs:
                api_kwargs["max_tokens"] = self.default_max_tokens
            # if input and input != "":
            #     api_kwargs["system"] = input
        else:
            raise ValueError(f"Model type {model_type} not supported")
        return api_kwargs

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
            BadRequestError,
        ),
        max_time=5,
    )
    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        """
        kwargs is the combined input and model_kwargs
        """
        if model_type == ModelType.EMBEDDER:
            raise ValueError(f"Model type {model_type} not supported")
        elif model_type == ModelType.LLM:
            return self.sync_client.messages.create(**api_kwargs)
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
            BadRequestError,
        ),
        max_time=5,
    )
    async def acall(
        self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED
    ):
        """
        kwargs is the combined input and model_kwargs
        """
        if self.async_client is None:
            self.async_client = self.init_async_client()
        if model_type == ModelType.EMBEDDER:
            raise ValueError(f"Model type {model_type} not supported")
        elif model_type == ModelType.LLM:
            return await self.async_client.messages.create(**api_kwargs)
        else:
            raise ValueError(f"model_type {model_type} is not supported")



from .generator_state_logger import GeneratorStateLogger
from .generator_call_logger import GeneratorCallLogger
from .decorators import trace_generator_states, trace_generator_call

__all__ = [
    "trace_generator_states",
    "trace_generator_call",
    "GeneratorStateLogger",
    "GeneratorCallLogger",
]



"""All data types used by Parameter, Optimizer, AdalComponent, and Trainer."""

from typing import List, Dict, Any
from enum import Enum
from dataclasses import dataclass, field
from datetime import datetime


from adalflow.core import DataClass


# TODO: set default optimization
class ParameterType(Enum):
    __doc__ = """Enum for the type of parameter to compute the loss with, and to inform the optimizer.

    The meaning of reach tuple is:
    1. First element: the name of the parameter.
    2. Second element: the description of the parameter.
    3. Third element: whether the parameter is trainable.

    To access each element, use the following:
    1. name: `ParameterType.PROMPT.value`
    2. description: `ParameterType.PROMPT.description`
    3. trainable: `ParameterType.PROMPT.default_trainable`
    """

    # trainable parameters with optimizers
    PROMPT = (
        "prompt",
        "Instruction to the language model on task, data, and format.",
        True,
    )  # optimized by tgd_optimizer
    DEMOS = (
        "demos",
        "A few examples to guide the language model.",
        True,
    )  # optimized by demo_optimizer

    # input and output parameters (similar to tensor, can have grad_opt true, but not trainable)
    INPUT = ("input", "The input to the component.", False)
    OUTPUT = ("output", "The output of the component.", True)
    HYPERPARAM = ("hyperparam", "Hyperparameters/args for the component.", False)

    # the following is a subtype of the output type
    # INSTANCE = ("instance", "Focus on fixing issues of this specific example.")
    GENERATOR_OUTPUT = (
        "generator_output",
        "The output of the generator.",
        True,
    )  # use raw response or error message as data, full response in full_response
    RETRIEVER_OUTPUT = ("retriever_output", "The output of the retriever.", True)
    LOSS_OUTPUT = ("loss", "The loss value.", True)
    SUM_OUTPUT = ("sum", "The sum of the losses.", True)
    NONE = ("none", "", False)

    def __init__(self, value: str, description: str, default_trainable: bool):
        self._value_ = value
        self.description = description
        self.default_trainable = default_trainable

    def __str__(self):
        """Return a string representation that includes the enum's value and description."""
        return f"{self.value} ({self.description})"

    def __repr__(self):
        """Return an unambiguous representation that is useful for debugging."""
        return f"<{self.__class__.__name__}.{self.name}: {self.value}, '{self.description}'>"


@dataclass
class EvaluationResult(DataClass):
    """A single evaluation of task pipeline response to a score in range [0, 1]."""

    score: float = field(
        default=0.0, metadata={"desc": "The score of the evaluation in range [0, 1]."}
    )
    feedback: str = field(
        default="",
        metadata={
            "desc": "Feedback on the evaluation, including reasons for the score."
        },
    )


@dataclass
class PromptData:
    id: str  # each parameter's id
    name: str  # each parameter's name
    data: str  # each parameter's data
    requires_opt: bool = field(
        default=True, metadata={"desc": "Whether this parameter requires optimization"}
    )


@dataclass
class TrainerStepResult(DataClass):
    step: int = field(default=0, metadata={"desc": "Step number"})
    val_score: float = field(
        default=None,
        metadata={
            "desc": "Validation score. Usually a smaller set than test set to chose the best parameter value."
        },
    )
    test_score: float = field(default=None, metadata={"desc": "Test score"})
    attempted_val_score: float = field(
        default=None, metadata={"desc": "Attempted validation score"}
    )
    prompt: List[PromptData] = field(
        default=None, metadata={"desc": "Optimized prompts for this step"}
    )


@dataclass
class TrainerValidateStats:
    """A single evaluation of task pipeline response to a score in range [0, 1]."""

    max_score: float = field(
        default=0.0, metadata={"desc": "The score of the evaluation in range [0, 1]."}
    )
    min_score: float = field(
        default=0.0, metadata={"desc": "The score of the evaluation in range [0, 1]."}
    )
    mean_of_score: float = field(
        default=0.0, metadata={"desc": "The score of the evaluation in range [0, 1]."}
    )
    std_of_score: float = field(
        default=0.0, metadata={"desc": "The score of the evaluation in range [0, 1]."}
    )


@dataclass
class TrainerResult(DataClass):
    steps: List[int] = field(default_factory=list, metadata={"desc": "List of steps"})
    val_scores: List[float] = field(
        default_factory=list, metadata={"desc": "List of validation scores"}
    )
    test_scores: List[float] = field(
        default_factory=list, metadata={"desc": "List of test scores"}
    )

    step_results: List[TrainerStepResult] = field(
        default_factory=list,
        metadata={"desc": "List of step results, in an aggregated form"},
    )
    effective_measure: Dict[str, Dict] = field(
        default_factory=dict,
        metadata={"desc": "Effective measures of the constrained training strategy"},
    )
    validate_stats: TrainerValidateStats = field(
        default=None,
        metadata={"desc": "Attempted Validation score statistics"},
    )
    time_stamp: str = field(
        default_factory=lambda: datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )
    total_time: float = field(
        default=0.0, metadata={"desc": "Total time taken for training"}
    )
    test_score: float = field(default=None, metadata={"desc": "Test score"})
    trainer_state: Dict[str, Any] = field(
        default=None, metadata={"desc": "Save the most detailed state of the trainer"}
    )



"""
Tool is LLM's extended capability which is one of the core design pattern of Agent. All tools can be wrapped in a FunctionTool class.
This helps to standardize the tool interface and metadata to communicate with the Agent.
"""

from typing import Any, Optional, Callable, Awaitable, Union
from inspect import iscoroutinefunction, ismethod, isfunction
import inspect
import logging
import asyncio
import nest_asyncio


from adalflow.core.types import (
    FunctionDefinition,
    FunctionOutput,
    Function,
)
from adalflow.core import Component
from adalflow.optim.parameter import Parameter
from adalflow.optim.grad_component import FunGradComponent
from adalflow.core.functional import (
    get_fun_schema,
)
from adalflow.utils import printc
from inspect import signature

AsyncCallable = Callable[..., Awaitable[Any]]

log = logging.getLogger(__name__)


def is_running_in_event_loop() -> bool:
    try:
        loop = asyncio.get_running_loop()
        if loop.is_running():
            return True
        else:
            return False
    except RuntimeError:
        return False


def find_instance_name_from_self(instance):
    """
    Attempt to find the variable name of the instance in the calling context.

    :param instance: The instance to find the name for.
    :return: The variable name of the instance, if found; otherwise, None.
    """
    # Inspect the calling stack frame
    frame = inspect.stack()[2].frame
    for var_name, var_obj in frame.f_locals.items():
        if var_obj is instance:
            return var_name
    return None


FunctionType = Union[Callable[..., Any], Awaitable[Callable[..., Any]]]


# TODO: improve the support for async functions, similarly a component might be used as a tool
class FunctionTool(Component):
    __doc__ = r"""Describing and Parsing(to LLM) and executing a function.

    Supports both normal callable functions and methods(__call__) of a component.
    When component is used, we support both the training and eval mode.

    Note:

        When the eval mode, it outputs FunctionOutput, and when the training mode, it outputs Parameter with data as FunctionOutput.

    Args:
        fn (Callable): The function to be executed.
        component (Component, optional): The component that owns the function. Defaults to None.
        definition (FunctionDefinition, optional): The definition of the function. Defaults to None.


    Function be used by LLM as a tool to achieve a specific task.

    What function can you pass as a tool?
    1. Any unbound function you wrote outside of a class.
    2. Any class method you wrote in your component. It can call `self` and other methods inside of your component.
    3. When the function is using a trainable component, and you can directly use the component's method as a tool or wrap it in a function. But you need to make sure to pass the component to the tool.

    Here are some examples:

    .. code-block:: python

        from adalflow.core.func_tool import FunctionTool
        class AgenticRAG(Component):
            def __init__(self, ...):
                super().__init__()
                self.retriever = Retriever()
                self.llm = Generator()

                def retriever_as_tool(input: str) -> str:
                    r"Used as a retriever tool."
                    return self.retriever(input)

                tools = [FunctionTool(retriever_as_tool, component=self.retriever),
                            FunctionTool(self.llm.__call__, component=self.llm)]
                # if you have trainable component, this will ensure it can be trained together with your whole task pipeline
                # if you dont want to train them and simply treating them as a tool, you can call like this
                # tools = [FunctionTool(retriever_as_tool), FunctionTool(self.llm.__call__, component=self.llm)]

    Features:

    - Supports both synchronous and asynchronous functions via `call` and `acall`.
    - Creates a `FunctionDefinition` from the function using `get_fun_schema`.
    - Executes the function with arguments.
        - Parses the function call expression (`FunctionExpression`) into `Function` (name, args, kwargs).
        - Executes the function using one of the following methods:
            - Via `call` with args and kwargs.
            - Via `eval`, without any context or sandboxing.
            - Via sandboxed execution directly using `sandbox_exec`.

    A FunctionTool allows other GradComponent(as a tool) to pass through correctly.
    """

    def __init__(
        self,
        fn: FunctionType,
        component: Optional[Component] = None,
        definition: Optional[FunctionDefinition] = None,
    ):
        super().__init__(
            name="FunctionTool", desc="A component calls and executes a function."
        )
        nest_asyncio.apply()
        assert fn is not None, "fn must be provided"

        # self.fn = fn  # it can be a function or component
        self.component = component  # pass it here to control the training mode
        if isinstance(fn, Component):
            self.fn = fn.__call__
        else:
            self.fn = fn
        self._is_async = iscoroutinefunction(fn)
        if isinstance(fn, Component):
            self.definition = (
                definition or self._create_fn_definition_for_grad_component(fn)
            )
        else:
            self.definition = definition or self._create_fn_definition()
        if self._is_async:
            log.info(f"FunctionTool: {fn} is async: {self._is_async}")

    @property
    def is_async(self) -> bool:
        return self._is_async

    def _create_fn_definition_for_grad_component(
        self, fn: FunGradComponent
    ) -> FunctionDefinition:
        name = fn.fun_name
        docstring = fn.doc_string
        signature_str = str(signature(fn.fun))
        instance = None
        return FunctionDefinition(
            func_name=name,
            func_desc=(
                f"{name}{signature_str}\nDocstring:{docstring}"
                if isinstance(docstring, str)
                else f"{name}{signature_str}\nDocstring:{docstring.data}"
            ),
            func_parameters=get_fun_schema(name, fn.fun),
            class_instance=instance,
        )

    def _create_fn_definition(self) -> FunctionDefinition:

        name = self.fn.__name__
        docstring = self.fn.__doc__
        signature_str = str(signature(self.fn))

        # Get the class that owns the method, if applicable
        cls_name = None
        instance = None
        if ismethod(self.fn):  # Check if it’s a bound method
            instance = self.fn.__self__
            instance = find_instance_name_from_self(instance)
            if name == "__call__" and not instance:
                raise ValueError(
                    "Please provide a name for the instance in the calling context"
                )
            cls_name = self.fn.__self__.__class__.__name__
        elif isfunction(self.fn):  # Unbound method
            cls_name = self.fn.__qualname__.split(".")[0]

        # Build the description
        description = f"{name}{signature_str}\n"
        if cls_name:
            description += f"Belongs to class: {cls_name}\n"
        if docstring:
            description += f"Docstring: {docstring}\n"

        # Get function parameters schema
        fn_parameters = get_fun_schema(name, self.fn)

        return FunctionDefinition(
            func_name=name,
            func_desc=description,
            func_parameters=fn_parameters,
            class_instance=instance,
        )

    def forward(self, *args, **kwargs) -> Parameter:
        r"""Forward the function tool."""
        return self.bicall(*args, **kwargs)

    def call(self, *args: Any, **kwargs: Any) -> FunctionOutput:
        r"""Execute the function synchronously.

        Example:

        .. code-block:: python

            import time
            def sync_function_1():
                time.sleep(1)
                return "Function 1 completed"

            tool_1 = FunctionTool(sync_function_1)
            output = tool_1.call()
        """
        return self.bicall(*args, **kwargs)
        # if self._is_async:
        #     raise ValueError("FunctionTool is asynchronous, use acall instead")
        # output, error = None, None
        # try:
        #     output = self.fn(*args, **kwargs)
        # except Exception as e:
        #     log.error(f"Error at calling {self.fn}: {e}")
        #     # raise ValueError(f"Error: {e}")
        #     error = str(e)
        # return FunctionOutput(
        #     name=self.definition.func_name,
        #     # raw_input={"args": args, "kwargs": kwargs},
        #     input=Function(name=self.definition.func_name, args=args, kwargs=kwargs),
        #     output=output,
        #     error=error,
        # )

    def bicall(self, *args: Any, **kwargs: Any) -> Union[FunctionOutput, Parameter]:
        r"""Execute the function synchronously.

        Example:

        .. code-block:: python

            import time
            def sync_function_1():
                time.sleep(1)
                return "Function 1 completed"

            tool_1 = FunctionTool(sync_function_1)
            output = tool_1.call()
        """
        if self._is_async:
            raise ValueError("FunctionTool is asynchronous, use acall instead")
        output, error = None, None

        # NOTE: special case:
        # self.fn can have both train and eval mode or untrainable as a function.
        try:
            # printc(f"args: {args}, kwargs: {kwargs}, fn: {self.fn}", color="yellow")
            output = self.fn(*args, **kwargs)
            # printc(f"output 1: {output}", color="yellow")
        except Exception as e:
            log.error(f"Error at calling {self.fn}: {e}")
            error = f"Error at calling {self.fn}: {e}"

        if isinstance(output, Parameter):
            if not self.training:
                raise ValueError(
                    f"FunctionTool {self.definition.func_name} is in eval mode, but the output is Parameter"
                )
            output.data = FunctionOutput(
                name=self.definition.func_name,
                # raw_input={"args": args, "kwargs": kwargs},
                input=Function(
                    name=self.definition.func_name, args=args, kwargs=kwargs
                ),
                output=output.data,
                error=error,
            )
            return output
        # printc(f"output: {output}", color="yellow")
        return FunctionOutput(
            name=self.definition.func_name,
            # raw_input={"args": args, "kwargs": kwargs},
            input=Function(name=self.definition.func_name, args=args, kwargs=kwargs),
            output=output,
            error=error,
        )

    async def acall(self, *args: Any, **kwargs: Any) -> FunctionOutput:
        r"""Execute the function asynchronously.

        Need to be called in an async function or using asyncio.run.

        Example:

        .. code-block:: python

            import asyncio
            async def async_function_1():
                await asyncio.sleep(1)  # Simulate async work
                return "Function 1 completed"

            async def call_async_function():
                tool_1 = FunctionTool(async_function_1)
                output = await tool_1.acall()

            asyncio.run(call_async_function())
        """
        if not self._is_async:
            raise ValueError("FunctionTool is not asynchronous, use call instead")
        output = None
        error = None
        try:
            output = await self.fn(*args, **kwargs)
        except Exception as e:
            log.error(f"Error at calling {self.fn}: {e}")
            error = str(e)

        return FunctionOutput(
            name=self.definition.func_name,
            input=Function(name=self.definition.func_name, args=args, kwargs=kwargs),
            output=output,
            error=error,
        )

    def execute(self, *args, **kwargs) -> FunctionOutput:
        r"""Execute the function synchronously or asynchronously based on the function type.

        No matter of the function type, you can run the function using both asyncio and without asyncio.


        Use it with caution as it might block the event loop.

        Example:

        .. code-block:: python

            import asyncio
            import time

            async def async_function_1():
                await asyncio.sleep(1)
                return "Function 1 completed"

            def sync_function_1():
                time.sleep(1)
                return "Function 1 completed"

            async def async_function_2():
                await asyncio.sleep(2)
                return "Function 2 completed"

            def sync_function_2():
                time.sleep(2)
                return "Function 2 completed"

            async_tool_1 = FunctionTool(async_function_1)
            sync_tool_1 = FunctionTool(sync_function_2)
            async_tool_2 = FunctionTool(async_function_2)
            sync_tool_2 = FunctionTool(sync_function_2)

            def run_sync_and_async_mix_without_wait():
                # both sync and async tool can use execute
                # sync tool can also use call
                # takes 5 seconds (1+1+2) + overhead
                start_time = time.time()
                results = [
                    async_tool_1.execute(),
                    sync_tool_1.execute(),
                    sync_tool_2.call(),
                ]
                end_time = time.time()
                print(f"run_sync_and_async_mix_without_wait time: {end_time - start_time}")
                return results

            async def run_sync_and_async_mix():
                # both sync and async tool can use execute&to_thread
                # async tool can also use acall without to_thread
                # takes a bit over 2 seconds max(2)
                start_time = time.time()
                results = await asyncio.gather(
                    async_tool_1.execute(),
                    sync_tool_1.execute(),
                    async_tool_2.acall(),
                )
                end_time = time.time()
                print(f"run_sync_and_async_mix time: {end_time - start_time}")
                return results

            run_sync_and_async_mix_without_wait()
            asyncio.run(run_sync_and_async_mix())
        """
        if self._is_async:
            log.debug(f"Running async function: {self.fn}")
            if is_running_in_event_loop():
                result = asyncio.create_task(self.acall(*args, **kwargs))
            else:
                result = asyncio.run(self.acall(*args, **kwargs))
        # NOTE: in juptyer notebook, it is always running in event loop
        else:
            log.debug(f"Running sync function: {self.fn}")
            if is_running_in_event_loop():
                log.debug(f"Running sync function in event loop: {self.fn}")
                result = asyncio.to_thread(self.call, *args, **kwargs)
            else:
                result = self.call(*args, **kwargs)

        return result

    # def __call__(self, *args, **kwargs) -> FunctionOutput:
    #     r"""Execute the function synchronously or asynchronously based on the function type."""
    #     return self.execute(*args, **kwargs)

    def _extra_repr(self) -> str:
        s = f"fn: {self.fn}, async: {self._is_async}, definition: {self.definition}"
        return s


if __name__ == "__main__":

    # import asyncio
    # import time

    # async def async_function_1():
    #     await asyncio.sleep(1)
    #     return "Function 1 completed"

    # def sync_function_1():
    #     time.sleep(1)
    #     return "Function 1 completed"

    # async def async_function_2():
    #     await asyncio.sleep(2)
    #     return "Function 2 completed"

    # def sync_function_2():
    #     time.sleep(2)
    #     return "Function 2 completed"

    # async_tool_1 = FunctionTool(async_function_1)
    # sync_tool_1 = FunctionTool(sync_function_2)
    # async_tool_2 = FunctionTool(async_function_2)
    # sync_tool_2 = FunctionTool(sync_function_2)

    # def run_sync_and_async_mix_without_wait():
    #     # both sync and async tool can use execute
    #     # sync tool can also use call
    #     # takes 5 seconds (1+1+2) + overhead
    #     start_time = time.time()
    #     results = [
    #         async_tool_1.execute(),
    #         sync_tool_1.execute(),
    #         sync_tool_2.call(),
    #     ]
    #     print(results)
    #     end_time = time.time()
    #     print(f"run_sync_and_async_mix_without_wait time: {end_time - start_time}")
    #     return results

    # async def run_sync_and_async_mix():
    #     # both sync and async tool can use execute&to_thread
    #     # async tool can also use acall without to_thread
    #     # takes a bit over 2 seconds max(2)
    #     start_time = time.time()
    #     results = await asyncio.gather(
    #         async_tool_1.execute(),
    #         sync_tool_1.execute(),
    #         async_tool_2.acall(),
    #     )
    #     print(results)
    #     end_time = time.time()
    #     print(f"run_sync_and_async_mix time: {end_time - start_time}")
    #     return results

    # print(async_tool_1.execute())

    # run_sync_and_async_mix_without_wait()
    # asyncio.run(run_sync_and_async_mix())

    from adalflow.components.model_client import OpenAIClient
    from adalflow.core.generator import Generator
    from adalflow.optim.parameter import Parameter
    from adalflow.core.types import GeneratorOutput
    from adalflow.utils import setup_env, printc

    setup_env()

    llm = Generator(
        model_client=OpenAIClient(),
        model_kwargs={"model": "gpt-3.5-turbo"},
    )
    # llm.train()

    def llm_as_tool(input: str, id: Optional[str] = None) -> str:
        """Used as a calculator tool."""
        printc(f"llm_as_tool: {input}", color="yellow")

        return llm(prompt_kwargs={"input_str": input}, id=id)

    llm_tool = FunctionTool(llm_as_tool, component=llm)
    llm_tool.train()
    output: Parameter = llm_tool("What is 2+2?")
    output.draw_graph()
    print(output)
    llm_tool.eval()
    output: FunctionTool = llm_tool("What is 2+2?")
    print(output)
    assert isinstance(output, FunctionOutput)
    assert isinstance(output.output, GeneratorOutput)

    # grad component

    from adalflow.optim.grad_component import fun_to_grad_component
    from adalflow.optim.parameter import ParameterType

    @fun_to_grad_component(
        desc="Finish",
        doc_string=Parameter(
            data="Finish the task with verbatim short factoid responses from retrieved context.",
            param_type=ParameterType.PROMPT,
            requires_opt=True,
            role_desc="Instruct how the agent creates the final answer from the step history.",
            name="doc_string",
        ),
    )
    def finish(answer: str, **kwargs) -> str:
        # """Finish the task with verbatim short factoid responses from retrieved context."""
        # printc(f"finish: {answer}", color="yellow")
        return answer

    finish_tool = FunctionTool(fn=finish, component=finish)

    definition = finish_tool.definition
    print(definition)
    # call function
    finish_tool.train()
    output: Parameter = finish_tool(
        "Finish the task with verbatim short factoid responses from retrieved context."
    )
    print(output)



import os
import sys


def get_adalflow_default_root_path() -> str:
    r"""This will be used for storing datasets, cache, logs, trained models, etc."""
    root = None
    if sys.platform == "win32":
        root = os.path.join(os.getenv("APPDATA"), "adalflow")
    else:
        root = os.path.join(os.path.expanduser("~"), ".adalflow")
    return root
